from OpenGL import GL
from PyQt5 import QtCore, QtWidgets
from numpy import pi, sin

from .Framework.Tools import glm
from .Framework.Tools.visualizationEnums import SegmentationTypes, ROIType, VisualizationActions, VisualizationObject
from .Framework.CoordinateSystem import CoordinateSystem
from .Framework.Tools.Camera import Camera

from .Framework.BoundingBox import BoundingBox
from .Framework.Bundle import Bundle
from .Framework.Mesh import Mesh
from .Framework.MRI import MRI
from .Framework.Slice import Slice
from .Framework.Volume import Volume

from .Framework.Segmentation.SegmentationHandler import SegmentationHandler
from .Framework.Segmentation.InPlaceSegmentation import InPlaceSegmentation
from .Framework.Segmentation.ROISegmentation import ROISegmentation
from .Framework.Segmentation.ROI import ROI
from .Framework.Segmentation.AtlasBasedParallelSegmentation import AtlasBasedParallelSegmentation
from .Framework.Segmentation.FFClustSegmentation import FFClustSegmentation

# Performance test
import numpy as np
import time as time


SCALE_FACTOR = 1/12000


def defaultLighting():
	return {
	"lightPos" : [0.0, 100.0, 0.0],
 	"lightAttr" : [0.5, 0.6, 1.0],
	"materialAttr" : [1.0, 0.8, 0.7],
	"shininess" : 5.0
	}


def defaultBackgroundColor():
	return [0.3, 0.3, 0.3, 1.0]


class GLWidget(QtWidgets.QOpenGLWidget):
	''' Widget with the logic of the program.
	It uses framework classes for visualization. And it manages all the
	framework instances.
	'''

	updateViewingObjects = QtCore.pyqtSignal()
	updateSettingWindow = QtCore.pyqtSignal(SegmentationHandler)
	objectsChanged = QtCore.pyqtSignal(list)

	def __init__(self, parent = None):
		''' Initialize the members for the OpenGL Widget
		'''

		super(GLWidget, self).__init__(parent)

		# self.version_profile = version_profile

		self.bundles = []
		self.meshes = []
		self.mris = []

		self.rois = []

		self.clearColor = defaultBackgroundColor()

		# default mouse flags
		self.orbit = False
		self.pan = False
		self.panning = False
		self.tilt = False

		self.currentObject = None
		self.draw_boundingbox = True

		# Mouse-screen attri
		self.screenX, self.screenY = 0,0

		self.segmentations = {	SegmentationTypes.InPlace : InPlaceSegmentation,
								SegmentationTypes.ROIs : ROISegmentation,
								SegmentationTypes.AtlasBased : AtlasBasedParallelSegmentation,
								SegmentationTypes.FFClust : FFClustSegmentation }

		self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu) # type: ignore

		self.customContextMenuRequested.connect(self.ShowContextMenu)


	def initializeGL(self):
		''' Initialize the OpenGL context and creates all the shaders programs
		'''

		print("OpenGL version: {}".format(GL.glGetString(GL.GL_VERSION).decode("utf-8"))) # type: ignore

		# Initialize shaders
		self.shaderDict = {}
		self.shaderList = [	Bundle.createProgram(),
							Mesh.createProgram(),
							Volume.createProgram(),
							BoundingBox.createProgram(),
							SegmentationHandler.createProgram(),
							# ROISegmentation.createProgram(),
							AtlasBasedParallelSegmentation.createProgram(),
							Slice.createProgram(),
							ROI.createProgram()]

		objectsList = [	[Bundle], #inplacesegmentation muved to segmentationhandler program
						[Mesh],
						[Volume],
						[BoundingBox],
						[InPlaceSegmentation, ROISegmentation],
						[AtlasBasedParallelSegmentation, FFClustSegmentation],
						[Slice],
						[ROI]]

		for i in range(len(objectsList)):
			self.shaderDict.update(dict.fromkeys(objectsList[i], self.shaderList[i]))

		self.coordSystemShader = CoordinateSystem.createProgram()

		self.camera = Camera(350.0)
		self.csCamera = Camera(1/sin(22.5*pi/180))

		# Config light parameters and view
		for s in self.shaderList:
			for shader in s:
				try:
					self._configLight(shader, defaultLighting())
				except:
					pass
				self._configView(shader)

		self._configView(self.coordSystemShader[0], self.csCamera)

		# Initialize some variables
		self.coordSystem = CoordinateSystem(self.coordSystemShader)
		self.proj = glm.perspective(45.0, 1, 0.01, 100000.0)
		self._configPerspective(self.coordSystemShader[0])


		self.testing = False
		self.FrameTime = GL.glGenQueries(1)

		##########
		print(GL.GL_MAX_UNIFORM_BLOCK_SIZE, GL.glGetIntegerv(GL.GL_MAX_UNIFORM_BLOCK_SIZE))
		print(GL.GL_MAX_VERTEX_UNIFORM_COMPONENTS, GL.glGetIntegerv(GL.GL_MAX_VERTEX_UNIFORM_COMPONENTS))
		print(GL.GL_MAX_TEXTURE_SIZE, ' {} max dim.'.format(GL.glGetIntegerv(GL.GL_MAX_TEXTURE_SIZE)))
		print(GL.GL_MAX_3D_TEXTURE_SIZE, ' {} max dim.'.format(GL.glGetIntegerv(GL.GL_MAX_3D_TEXTURE_SIZE)))



		GL.glEnable(GL.GL_DEPTH_TEST)
		GL.glEnable(GL.GL_PRIMITIVE_RESTART)
		GL.glPrimitiveRestartIndex(0xFFFFFFFF)


	def paintGL(self):
		''' Paint function. Its called by the widget.
		'''

		GL.glViewport(0, 0, self.width(), self.height())

		GL.glClearColor(*self.clearColor)
		GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)

		# FRAME TIME
		if self.testing:
			GL.glBeginQuery(GL.GL_TIME_ELAPSED, self.FrameTime)

		for bundle in self.bundles:
			bundle.draw()

		for mri in self.mris:
			mri.draw()

		for roi in self.rois:
			roi.draw()

		for mesh in self.meshes:
			mesh.draw(self.camera.eye)

		GL.glClear(GL.GL_DEPTH_BUFFER_BIT)
		GL.glViewport(0, 0, 150, 150)
		self.coordSystem.draw()

		# FRAME TIME
		if self.testing:
			GL.glEndQuery(GL.GL_TIME_ELAPSED)  # Termina consulta.
			ready = False
			while not ready:
				ready = GL.glGetQueryObjectuiv(self.FrameTime, GL.GL_QUERY_RESULT_AVAILABLE)

			self.testArray[self.testing_i] = (GL.glGetQueryObjectuiv(
				self.FrameTime, GL.GL_QUERY_RESULT)/1000000000.0)
			self.testing_i += 1


	def resizeGL(self, width, height):
		''' Recalculate the projection matrix, and upload the new one to all shaders
		'''

		self.proj = glm.perspective(45.0, width/height, 0.01, 100000.0)
		for s in self.shaderList:
			for shader in s:
				self._configPerspective(shader)


	def validBundleExtension(self):
		return Bundle.validExtension()


	def validMeshExtension(self):
		return Mesh.validExtension()


	def validMRIExtension(self):
		return MRI.validExtension()


	def addBundleFile(self, bundleFiles):
		# Makes the context the current one
		self.makeCurrent()

		new_bundles = [Bundle(bundlePath, self.shaderDict, self.bundles) for bundlePath in bundleFiles]
		if new_bundles:
			self.bundles.extend(new_bundles)

			if hasattr(self, 'camera'):
				bboxs = np.array([bundle.calculateBoundingBoxDimCenter() for bundle in new_bundles])
				center = (bboxs[:, 0].max(axis=0) + bboxs[:, 0].min(axis=0)) / 2

				self.camera.center = center
				self.camera.view = self.camera.calculateView()

				for s in self.shaderList:
					for shader in s:
						self._configView(shader)

			self._configView(self.coordSystemShader[0], self.csCamera)

		self.updateViewingObjects.emit()
		self.update()


	def addMeshFile(self, meshFiles):
		# Makes the context the current one
		self.makeCurrent()

		for meshPath in meshFiles:
			self.meshes.append(Mesh(meshPath, self.shaderDict, self.meshes))
			# self.meshes.append(Mesh(meshPath, self.meshShader, self.meshes))

		self.updateViewingObjects.emit()
		self.update()


	def addMRIFile(self, mriFiles):
		# Makes the context the current one
		self.makeCurrent()

		for mriPath in mriFiles:
			self.mris.append(MRI(mriPath, self.shaderDict, self.mris))

		self.updateViewingObjects.emit()
		self.update()


	def addROI(self):
		self.makeCurrent()

		self.rois.append(ROI(ROIType.Sphere, self.camera.center, 30.0, self.shaderDict, self.rois))

		self.updateViewingObjects.emit()
		self.update()


	@QtCore.pyqtSlot(dict)
	def modifyObject(self, options):
		# Makes the context the current one
		self.makeCurrent()

		def retrieveItemIteratively(root, index):
			item = root

			for i in index:
				if hasattr(item, 'children'):
					item = item.children[i]
				else:
					item = item[i]

			return item

		objectRoot = None
		itemObject = None

		rootType = options['rootType']
		index = options['number']
		action = options['action']

		if rootType == VisualizationObject.Bundle:
			objectRoot = self.bundles
		elif rootType == VisualizationObject.Mesh:
			objectRoot = self.meshes
		elif rootType == VisualizationObject.MRI:
			objectRoot = self.mris
		elif rootType == VisualizationObject.ROI:
			objectRoot = self.rois
		else:
			raise TypeError('Unrecognise visualization objects type: ', rootType)

		if index != -1:
			itemObject = retrieveItemIteratively(objectRoot, index)


		# Delete
		#########################
		if action == VisualizationActions.Delete:
			if index != -1:
				itemObject.cleanOpenGL()
				itemObject.parent.remove(itemObject)
				itemObject.clearRef()
			else:
				while(len(objectRoot) != 0):
					item = objectRoot[0]
					item.cleanOpenGL()
					objectRoot.remove(item)
					item.clearRef()
			self.updateViewingObjects.emit()

		# Toggle drawable
		#########################
		elif action == VisualizationActions.ToggleDrawable:
			itemObject.drawable = not itemObject.drawable

		# Rotate
		#########################
		elif action == VisualizationActions.Rotate:
			angle, axis = options['data']

			if index == -1:
				for i in objectRoot:
					i.rotate(self.camera.center, angle, axis)
				self.objectsChanged.emit(objectRoot)

			else:
				itemObject.rotate(self.camera.center, angle, axis)
				self.objectsChanged.emit([itemObject])

		# Translate
		#########################
		elif action == VisualizationActions.Translate:
			v = options['data']

			if index == -1:
				for i in objectRoot:
					i.translate(v)
				self.objectsChanged.emit(objectRoot)
			else:
				itemObject.translate(v)
				self.objectsChanged.emit([itemObject])

		# Scale
		#########################
		elif action == VisualizationActions.Scale:
			v = options['data']

			if index == -1:
				for i in objectRoot:
					i.scale(v, self.camera.center)
				self.objectsChanged.emit(objectRoot)

			else:
				itemObject.scale(v, self.camera.center)

				self.objectsChanged.emit([itemObject])

		# ResetTransforms
		##########################}
		elif action == VisualizationActions.ResetTransforms:
			if index == -1:
				for i in objectRoot:
					i.resetModel()
				self.objectsChanged.emit(objectRoot)

			else:
				itemObject.resetModel()

				self.objectsChanged.emit([itemObject])

		# Load and Apply Matrix
		elif action == VisualizationActions.LoadAndApplyMatrix:
			matrixFile = options['data']

			itemObject.loadAndApplyMatrix(matrixFile)

		# Add Segmentation
		# ########################
		elif action == VisualizationActions.AddSegmentation:
			segmentationType = options['data']

			if segmentationType == SegmentationTypes.InPlace:
				segmentationObject = InPlaceSegmentation(itemObject, self.shaderDict)

			elif segmentationType == SegmentationTypes.ROIs:
				segmentationObject = ROISegmentation(itemObject, self.shaderDict)

			elif segmentationType == SegmentationTypes.AtlasBased:
				segmentationObject = AtlasBasedParallelSegmentation(itemObject, self.shaderDict)

			elif segmentationType == SegmentationTypes.FFClust:
				segmentationObject = FFClustSegmentation(itemObject, self.shaderDict)

			# Volume visualizations
			# #####################
			elif segmentationType == VisualizationObject.MRISlice:
				segmentationObject = Slice(itemObject, self.shaderDict)

			elif segmentationType == VisualizationObject.MRIVolume:
				segmentationObject = Volume(itemObject, self.shaderDict, self.camera)


			else:
				raise TypeError('Segmentation type not implemented yet: ', segmentationType)

			itemObject.children.append(segmentationObject)
			self.updateViewingObjects.emit()

		# Slice Modification
		elif action == VisualizationActions.SliceModification:
			linearInterp, discardValues, threshold, axis, pos2Axis, bright, contrast = options['data']

			itemObject.setLinearInterpolation(linearInterp)
			itemObject.setDiscardValuesWithThreshold(discardValues, threshold)
			itemObject.calculatePlaneNormal(axis, pos2Axis)
			itemObject.setBright(bright)
			itemObject.setContrast(contrast)

		# Mesh Modification
		elif action == VisualizationActions.MeshModification:
			if len(options['data']) == 1:
				if isinstance(options['data'][0], bool):
					itemObject.setFront2BackSorting(options['data'][0])
				else:
					itemObject.setColor(options['data'][0])

			else:
				drawLines, alpha = options['data']

				if index == -1:
					for i in objectRoot:
						i.setDrawLines(drawLines)
						i.setAlpha(alpha)
				else:
					itemObject.setDrawLines(drawLines)
					itemObject.setAlpha(alpha)

		# Volume Modification
		elif action == VisualizationActions.VolumeModification:
			data = options['data']
			if isinstance(data, bool):
				itemObject.setFront2BackDrawing(data)
			elif len(data) == 2:
				itemObject.setThreshold(data[0])
				itemObject.setAlpha(data[1])
			else:
				raise TypeError('Unrecognise data')

		# Shader selection
		elif action == VisualizationActions.ShaderSelection:
			data = options['data']

			if index == -1:
				for i in objectRoot:
					i.setSelectedShader(data)
			else:
				itemObject.setSelectedShader(data)
		# Focus camera on object
		elif action == VisualizationActions.FocusObject:
			size = None
			center = itemObject.get_center()
			try:
				size = np.max(itemObject.get_size()) * 2
			except NotImplementedError:
				print(f"Object {itemObject} does not implement get_size.")
			finally:
				self.camera.focus(center, min_r=size)
				for shaders in self.shaderList:
					for shader in shaders:
						self._configView(shader)
		else:
			raise TypeError('Unrecognise visualization objects action')

		self.update()


	@QtCore.pyqtSlot(dict)
	def selectedObject(self, options):
		self.currentObject = options.get('current')

	@QtCore.pyqtSlot(dict)
	def modifySegmentation(self, options):
		# Makes the context the current one
		self.makeCurrent()

		segmentation = options.get('reference')
		attributes = options.get('attributes')
		segmentSubject = options.get('segmentSubject')
		updateViewingObjects = options.get('updateViewingObjects')
		updateSettingWindow = options.get('updateSettingWindow')

		if not isinstance(segmentation, SegmentationHandler):
			raise TypeError('modfySegmentation got an unxpected object reference: ', type(segmentation))

		if attributes:
			for attr in attributes.keys():
				try:
					getattr(segmentation, attr)(*attributes[attr])
				except Exception as e:
					QtWidgets.QMessageBox.warning(self, "Error", str(e), QtWidgets.QMessageBox.Ok)

		if segmentSubject:
			segmentation.segmentSubject()

		if updateViewingObjects:
			self.updateViewingObjects.emit()

		if updateSettingWindow:
			self.updateSettingWindow.emit(segmentation)

		self.update()


	def _configLight(self, shader, attributes):
		lightPos = attributes['lightPos']
		light = attributes['lightAttr']
		lightLa = light[0]
		lightLd = light[1]
		lightLs = light[2]
		material = attributes['materialAttr']
		materialKa = material[0]
		materialKd = material[1]
		materialKs = material[2]
		shininess = attributes['shininess']

		shader.glUseProgram()

		# Light
		GL.glUniform4f(shader.glGetUniformLocation("Light.pos"), *lightPos, 1.0)
		GL.glUniform3f(shader.glGetUniformLocation("Light.La"), lightLa, lightLa, lightLa)
		GL.glUniform3f(shader.glGetUniformLocation("Light.Ld"), lightLd, lightLd, lightLd)
		GL.glUniform3f(shader.glGetUniformLocation("Light.Ls"), lightLs, lightLs, lightLs)

		# Material
		GL.glUniform3f(shader.glGetUniformLocation("Material.Ka"), materialKa, materialKa, materialKa)
		GL.glUniform3f(shader.glGetUniformLocation("Material.Kd"), materialKd, materialKd, materialKd)
		GL.glUniform3f(shader.glGetUniformLocation("Material.Ks"), materialKs, materialKs, materialKs)
		GL.glUniform1f(shader.glGetUniformLocation("Material.shininess"), shininess)


	def _configPerspective(self, shader):
		shader.glUseProgram()
		GL.glUniformMatrix4fv(shader.glGetUniformLocation("P"),
		                      1, GL.GL_TRUE, self.proj.getA())

	def _configView(self, shader, camera=None):
		if camera == None:
			camera = self.camera
		shader.glUseProgram()
		GL.glUniformMatrix4fv(shader.glGetUniformLocation("V"),
		                      1, GL.GL_FALSE, camera.view.getA())



	def mousePressEvent(self, event):
		""" Captura el evento del boton del mouse clickeado y selecciona la funcion correspondiente
			a este. """
		self.screenX, self.screenY = event.pos().x(), event.pos().y()

		if event.button() == QtCore.Qt.LeftButton:
			self.orbit = True
			self.pan = False
			self.tilt = False
			event.accept()

		elif event.button() == QtCore.Qt.RightButton: # Permite desplazar el modelo.
			self.orbit = False
			self.pan = True
			self.panning = False
			self.tilt = False
			event.accept()

		elif event.button() == QtCore.Qt.MiddleButton:
			self.orbit = False
			self.pan = False
			self.tilt = True
			event.accept()

		else:
			self.orbit = False
			self.pan = False
			self.tilt = False
			event.ignore()


	def mouseMoveEvent(self, event):
		""" Ejecuta la funcion correspondiente al boton del mouse clickeado. """
		newScreenX, newScreenY = event.pos().x(), event.pos().y()
		dx, dy = newScreenX - self.screenX, newScreenY - self.screenY

		self.screenX, self.screenY = newScreenX, newScreenY

		if QtWidgets.QApplication.keyboardModifiers() & QtCore.Qt.ControlModifier:
			if not self.currentObject:
				return

			if self.orbit:
				axis, angle = self.camera.axisAndAngleFromScreen(dx, dy)

				if isinstance(self.currentObject, list):
					for cObject_i in self.currentObject:
						cObject_i.rotate(cObject_i.get_center(), angle, axis)
				else:
					self.currentObject.rotate(self.currentObject.get_center(), angle, axis)

			elif self.pan:
				self.panning = True
				translateV = self.camera.vectorFromScreen(dx, dy)

				if isinstance(self.currentObject, list):
					for cObject_i in self.currentObject:
						cObject_i.stackTranslate(translateV)
				else:
					self.currentObject.stackTranslate(translateV)
			elif self.tilt:
				axis = self.camera.tiltAxisFromScreen()
				if isinstance(self.currentObject, list):
					for cObject_i in self.currentObject:
						cObject_i.spin(dx * np.pi / 10, axis)
				else:
					self.currentObject.spin(dx *np.pi / 10, axis)
			self.objectsChanged.emit([self.currentObject])
		else:
			if self.orbit:
				if True:
					self.camera.orbit(dx, dy)
					self.csCamera.orbit(dx, dy)
				self._configView(self.coordSystemShader[0], self.csCamera)
			elif self.pan:
				self.panning = True
				self.camera.panning(dx, dy)
			elif self.tilt:
				self.camera.spin(np.pi * dx / 10)
				self.csCamera.spin(np.pi * dx / 10)
				self._configView(self.coordSystemShader[0], self.csCamera)
			else:
				return

			for s in self.shaderList:
				for shader in s:
					self._configView(shader)

		event.accept()
		self.update()


	def wheelEvent(self, event):
		if QtWidgets.QApplication.keyboardModifiers() & QtCore.Qt.ControlModifier:
			if not self.currentObject:
				return

			delta = event.angleDelta().y()

			if isinstance(self.currentObject, list):
				for cObject_i in self.currentObject:
					cObject_i.stackScale(delta*SCALE_FACTOR+1.0, self.camera.center)
			else:
				self.currentObject.stackScale(delta*SCALE_FACTOR+1.0, self.camera.center)

			self.objectsChanged.emit([self.currentObject])

		else:
			self.camera.zooming(event.angleDelta().y())

			for s in self.shaderList:
				for shader in s:
					self._configView(shader)

		event.accept()
		self.update()

	def set_draw_bbox(self, value=None):
		if value is None:
			self.draw_boundingbox = not self.draw_boundingbox
		else:
			self.draw_boundingbox = value
		for bundle in self.bundles:
			bundle.setDrawBB(self.draw_boundingbox)
		for mri in self.mris:
			mri.setDrawBB(self.draw_boundingbox)
		for mesh in self.meshes:
			mesh.setDrawBB(self.draw_boundingbox)


	def keyPressEvent(self, event):
		key = event.key()

		if key == QtCore.Qt.Key_1:
			self.camera.frontView()
			self.csCamera.frontView()
		elif key == QtCore.Qt.Key_2:
			self.camera.backView()
			self.csCamera.backView()
		elif key == QtCore.Qt.Key_3:
			self.camera.leftView()
			self.csCamera.leftView()
		elif key == QtCore.Qt.Key_4:
			self.camera.rightView()
			self.csCamera.rightView()
		elif key == QtCore.Qt.Key_5:
			self.camera.topView()
			self.csCamera.topView()
		elif key == QtCore.Qt.Key_6:
			self.camera.bottomView()
			self.csCamera.bottomView()
		elif key == QtCore.Qt.Key_B:
			self.set_draw_bbox()
		elif key == QtCore.Qt.Key_W:
			self.camera.panning(0, 0, -10)
		elif key == QtCore.Qt.Key_A:
			self.camera.panning(10, 0)
		elif key == QtCore.Qt.Key_S:
			self.camera.panning(0, 0, 10)
		elif key == QtCore.Qt.Key_D:
			self.camera.panning(-10, 0)
		elif key == QtCore.Qt.Key_R:
			self.camera.panning(0, 10)
		elif key == QtCore.Qt.Key_F:
			self.camera.panning(0, -10)
		elif key == QtCore.Qt.Key_Q:
			self.camera.spin(np.pi * 1)
			self.csCamera.spin(np.pi * 1)
		elif key == QtCore.Qt.Key_E:
			self.camera.spin(np.pi * -1)
			self.csCamera.spin(np.pi * -1)
		elif key == QtCore.Qt.Key_Up:
			self.camera.orbit(0, 1)
			self.csCamera.orbit(0, 1)
		elif key == QtCore.Qt.Key_Down:
			self.camera.orbit(0, -1)
			self.csCamera.orbit(0, -1)
		elif key == QtCore.Qt.Key_Left:
			self.camera.orbit(1, 0)
			self.csCamera.orbit(1, 0)
		elif key == QtCore.Qt.Key_Right:
			self.camera.orbit(-1, 0)
			self.csCamera.orbit(-1, 0)
		elif key == QtCore.Qt.Key_F4:
			self.debug_breakpoint()
		else:
			event.ignore()
			return

		for s in self.shaderList:
			for shader in s:
				self._configView(shader)

		self._configView(self.coordSystemShader[0], self.csCamera)


		event.accept()
		self.update()


	@QtCore.pyqtSlot(QtCore.QPoint)
	def ShowContextMenu(self, pos):
		if self.panning:
			return

		self.makeCurrent()

		contextMenu = QtWidgets.QMenu()

		resetCameraAction = contextMenu.addAction("Reset camera")

		action = contextMenu.exec(self.mapToGlobal(pos))

		if action == resetCameraAction:
			self.camera.defaultValues()
			self.csCamera.defaultValues()

			for s in self.shaderList:
				for shader in s:
					self._configView(shader)

			self._configView(self.coordSystemShader[0], self.csCamera)

		self.update()



	def testingModule(self):
		print('testingModule')
		self.segmentationTest()


	def cleanOpenGL(self):
		self.makeCurrent()

		for bundle in self.bundles:
			bundle.cleanOpenGL()

		for mesh in self.meshes:
			mesh.cleanOpenGL()

		for mri in self.mris:
			mri.cleanOpenGL()

		for roi in self.rois:
			roi.cleanOpenGL()

		self.coordSystem.cleanOpenGL()


	def performanceTestFPS(self):
		self.makeCurrent()
		testingN = 1000
		n = 100
		self.testing_i = 0
		self.testing = True
		self.testArray = np.empty(testingN*n, dtype=np.float32)
		for i in range(testingN):
			self.camera.orbit(10, 0)
			self.csCamera.orbit(10, 0)

			self._configView(self.coordSystemShader[0], self.csCamera)

			for s in self.shaderList:
				for shader in s:
					self._configView(shader)

			for j in range(n):
				self.paintGL()

		self.testing = False

		print("Test ended. Number of samples taken: ", self.testing_i, "\nAverage frame time:", np.mean(self.testArray), "\nAverage fps: ", 1/np.mean(self.testArray))


	def segmentationTest(self):
		self.makeCurrent()
		try:
			testingN = 1000
			n = 10
			self.testing_i = 0
			self.testArray = np.empty(testingN*n, dtype=np.float32)

			for i in range(testingN):
				for j in range(n):
					tb = time.time()
					self.bundles[0].children[0].segmentSubject()
					te = time.time()

					self.testArray[self.testing_i] = te-tb
					self.testing_i += 1

			print("Test ended. Number of samples taken: ", self.testing_i, "\nAverage time:", np.mean(self.testArray))
		except Exception as e:
			print(e)

	def debug_breakpoint(self):
		if not self.currentObject:
			return
		self.objectsChanged.emit([self.currentObject])

		for s in self.shaderList:
			for shader in s:
				self._configView(shader)


##
# @mainpage Phybers
# @brief Phybers is a library for Python that integrates multiple tractography and neural-fibers related algorithms. It includes the following sub-packages:
#
# \tableofcontents
# @subsection Modules Modules:
# <ol>
# <li>\ref phybers.clustering.clustering "Clustering" </li>
# <li>\ref phybers.segment.segment "Segment"</li>
# <li>\ref phybers.utils.utilities "Utilities"</li>
# <li>\ref phybers.fibervis "Fibervis"</li>
# </ol>
# Here goes a footer of the package or something



from PyQt5 import QtGui, QtWidgets, uic, QtCore
from functools import wraps
from importlib_resources import files
from ...Framework.VisualizationBaseObject import VisualizationObject

_roisd_ui = files('phybers.fibervis.ui.Segmentations').joinpath(
	'ROIsSegmentationDialog.ui')


# MUST DELETE
def identifyNumberRecursively(item, number):
	root = item

	if item.parent() != None:
		root = identifyNumberRecursively(item.parent(), number)
		number.append(item.parent().indexOfChild(item))
	else:
		return item
	return root


def retrieveItemIteratively(root, index):
	item = root

	for i in index:
		if hasattr(item, 'children'):
			item = item.children[i]
		else:
			item = item[i]

	return item
	###########


def detectAlways(func):
	@wraps(func)
	def wrapper(*args, **kwargs):
		func(*args, **kwargs)

		if args[0].ui.alwaysBox.isChecked():
			args[0].segmentSubject()
	return wrapper


class ROIsSegmentationDialog(QtWidgets.QDialog):
	updateObject = QtCore.pyqtSignal(dict)


	def __init__(self, segmentation, parent):
		super().__init__(parent)
		self.ui = uic.loadUi(str(_roisd_ui), self)

		self.segmentationRef = segmentation

		self.ui.tableWidget.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
		self.ui.tableWidget.horizontalHeader().setStretchLastSection(True)

		self.ui.addROIButton.clicked.connect(self.addROI)
		self.ui.removeROIButton.clicked.connect(self.removeROI)

		self.ui.detectButton.clicked.connect(self.segmentSubject)

		self.ui.exportButton.clicked.connect(self.exportBundle)

		self.ui.alphaSlider.valueChanged.connect(self.alphaChanged)

		self.ui.tableWidget.cellChanged.connect(self.cellChanged)

		self.ui.logicLineEdit.editingFinished.connect(self.logicLineEditEdited)

		self.show()




	# // Actualiza dibujo al clickear checkbox de la tabla
	# this.connect(this.ui.tableWidget, SIGNAL(cellChanged(int,int)), this, SLOT(itemtoggle(int,int)));


	@detectAlways
	@QtCore.pyqtSlot()
	def addROI(self):
		treeWidgetItems = self.parentWidget().ui.treeWidget.selectedItems()
		if len(treeWidgetItems) == 0:
			return
		elif treeWidgetItems[0].identifier != VisualizationObject.ROI:
			QtWidgets.QMessageBox.question(self, "Invalid ROI", 'Must select a ROI from the visualization tool dialog.', QtWidgets.QMessageBox.Ok)
			return

		# Bad practice
		##############
		number = []
		root = identifyNumberRecursively(treeWidgetItems[0], number)

		if root.text(0) == 'Bundles':
			reference = self.parentWidget().bundleReference
		elif root.text(0) == 'Meshes':
			reference = self.parentWidget().meshReference
		elif root.text(0) == 'MRIs':
			reference = self.parentWidget().mriReference
		elif root.text(0) == 'ROIs':
			reference = self.parentWidget().roisReference
		else:
			reference = None

		refObj = retrieveItemIteratively(reference, number)
		if isinstance(refObj, list):
			QtWidgets.QMessageBox.question(self, "Too many ROIs", 'Please select one ROI.', QtWidgets.QMessageBox.Ok)
			return
		##############

		options = {	'reference' : self.segmentationRef,
					'attributes': {'addROI' : [refObj]},
					'updateViewingObjects' : True}

		self.updateObject.emit(options)


	@detectAlways
	@QtCore.pyqtSlot()
	def removeROI(self):
		index = self.ui.tableWidget.currentRow()

		# self.ui.tableWidget.removeRow(index)
		options = {	'reference' : self.segmentationRef,
					'attributes': {'removeROIFromIndex' : [index]},
					'updateViewingObjects' : True}

		self.updateObject.emit(options)


	@QtCore.pyqtSlot()
	def segmentSubject(self):
		options = {	'reference' : self.segmentationRef,
					'segmentSubject' : True}

		self.updateObject.emit(options)


	@QtCore.pyqtSlot()
	def alphaChanged(self):
		alpha = self.alphaSlider.value()/(self.alphaSlider.maximum()+1)

		options = {	'reference' : self.segmentationRef,
					'attributes': {'setAlpha' : [alpha]}}

		self.updateObject.emit(options)


	@QtCore.pyqtSlot()
	def exportBundle(self):

		# Gets direction of bundle files
		outFile, fileType = QtWidgets.QFileDialog.getSaveFileName(self, "Save Bundle file", "", "Bundle (*.bundles)")#;;Trk file (*.trk)")

		options = {	'reference' : self.segmentationRef,
					'attributes' : {'exportFile' : [outFile]}}

		self.updateObject.emit(options)


	@detectAlways
	@QtCore.pyqtSlot(int, int)
	def cellChanged(self, row, column):
		if column == 0:
			checked = self.ui.tableWidget.item(row, column).checkState() == QtCore.Qt.Checked

			options = {	'reference' : self.segmentationRef,
						'attributes' : {'setValidatorAtIndex' : [row, checked]}}

			self.updateObject.emit(options)


	@detectAlways
	@QtCore.pyqtSlot()
	def logicLineEditEdited(self):
		options = {	'reference' : self.segmentationRef,
					'attributes': {'updateLogic' : [str(self.ui.logicLineEdit.text()),
				     								self.ui.logicLineEdit.setText]}}

		self.updateObject.emit(options)


	def update(self):
		while self.ui.tableWidget.rowCount():
			self.ui.tableWidget.removeRow(0)

		for i in range(len(self.segmentationRef.children)):
			self.insertItemInTable(self.segmentationRef.children[i], self.segmentationRef.roiValidator[i])

	def objectsChanged(self, objects):
		if self.ui.alwaysBox.isChecked():
			for obj in objects:
				if obj in self.segmentationRef.children or obj is self.segmentationRef:
					self.segmentSubject()
					return


	def insertItemInTable(self, objRef, checked):
		self.ui.tableWidget.cellChanged.disconnect(self.cellChanged)

		row = self.ui.tableWidget.rowCount()
		self.ui.tableWidget.insertRow(row)

		tag = QtWidgets.QTableWidgetItem(str(row))
		tag.setCheckState(QtCore.Qt.Checked if checked else QtCore.Qt.Unchecked)
		tag.setTextAlignment(QtCore.Qt.AlignHCenter)
		tag.setFlags(tag.flags() & ~QtCore.Qt.ItemIsEditable)

		bg = QtWidgets.QTableWidgetItem()
		bg.setBackground(QtGui.QBrush(QtGui.QColor(
						int(objRef.getRGB256()[0]),
						int(objRef.getRGB256()[1]),
						int(objRef.getRGB256()[2]))))
		bg.setFlags(bg.flags() & ~QtCore.Qt.ItemIsEditable)

		roiType = QtWidgets.QTableWidgetItem(str(objRef.roiType))
		roiType.setTextAlignment(QtCore.Qt.AlignHCenter)
		roiType.setFlags(roiType.flags() & ~QtCore.Qt.ItemIsEditable)

		self.ui.tableWidget.setItem(row, 0, tag)
		self.ui.tableWidget.setItem(row, 1, bg)
		self.ui.tableWidget.setItem(row, 2, roiType)

		self.ui.tableWidget.selectRow(row)

		self.ui.tableWidget.cellChanged.connect(self.cellChanged)



'''
InPlaceSegmentationDialog.exportBundle ----> trk not implemented
'''

from .GLContext import GLWidget
from PyQt5 import QtGui, QtWidgets, uic, QtCore
from .ui.Controllers.VisDialog import VisDialog
from .Framework.Tools.performance import timeit
from importlib_resources import files

_viewer = files('phybers.fibervis.ui').joinpath('viewer.ui')

class WindowController(QtWidgets.QMainWindow):
	''' Window controller for the app. It has function for several action.
	'''
	def __init__(self):
		super().__init__()

		glFormat = QtGui.QSurfaceFormat()
		glFormat.setVersion(4, 1)
		glFormat.setProfile(QtGui.QSurfaceFormat.CoreProfile)
		glFormat.setSamples(4)

		QtGui.QSurfaceFormat.setDefaultFormat(glFormat)

		# versionProfile = QtGui.QOpenGLVersionProfile()
		# versionProfile.setVersion(4, 1)
		# versionProfile.setProfile(QtGui.QSurfaceFormat.CoreProfile)

		# self.contextHandler = GLContext.GLWidget(versionProfile, self)
		self.contextHandler = GLWidget(self)
		self.contextHandler.setFocusPolicy(QtCore.Qt.StrongFocus)

		self.ui = uic.loadUi(_viewer, self)
		self.visObj = VisDialog(self)

		self.ui.setCentralWidget(self.contextHandler)

		self.ui.setWindowTitle('leFiber v1.0.0')

		# Connection of GUI to GLCBontext
		self.ui.actionOpenBundle.triggered.connect(self.openBundleFile)
		self.ui.actionOpenMesh.triggered.connect(self.openMeshFile)
		self.ui.actionOpenMRI.triggered.connect(self.openMRIFile)
		self.ui.actionAddROISphere.triggered.connect(self.createROI)

		#Toggle Bounding Boxes
		self.ui.actionBoundingBoxes.changed.connect(
			lambda: self.contextHandler.set_draw_bbox(self.ui.actionBoundingBoxes.isChecked()))
		self.ui.actionBoundingBoxes.setChecked(True)

		# testing action
		self.ui.actionTest.triggered.connect(self.contextHandler.testingModule)

		# Visualization objects window
		# changed
		# hovered
		# toggled
		self.ui.actionVisualizationObjects.changed.connect(self.visObj.toggleVisWin(self.ui.actionVisualizationObjects.isChecked, self.pos, self.size))

		# Closed visualization win
		self.visObj.closed.connect(self.unCheckActionVisualizationObjects)

		# New visualization object
		self.contextHandler.updateViewingObjects.connect(self.visObj.updateTree)
		self.contextHandler.updateSettingWindow.connect(self.visObj.updateSettingWindow)

		# Modify visualization object
		self.visObj.changeObject.connect(self.contextHandler.modifyObject)
		self.visObj.modifySegmentation.connect(self.contextHandler.modifySegmentation)
		self.visObj.selectedObject.connect(self.contextHandler.selectedObject)

		# Notify R, T and S on objects
		self.contextHandler.objectsChanged.connect(self.visObj.objectsChanged)

		# We set the reference for the visualization objects
		self.visObj.bundleReference = self.contextHandler.bundles
		self.visObj.meshReference = self.contextHandler.meshes
		self.visObj.mriReference = self.contextHandler.mris
		self.visObj.roisReference = self.contextHandler.rois

		self.setAcceptDrops(True)
		# Hide testing tab
		self.menuTesting.menuAction().setVisible(False)

	def dragEnterEvent(self, event):
		if (event.mimeData().hasUrls()):
			event.acceptProposedAction()

	@timeit
	def dropEvent(self, event):
		for file in event.mimeData().urls():
			fileName = file.toLocalFile()

			extension = fileName.split('.')[-1]

			# Bundle
			if extension in self.contextHandler.validBundleExtension():
				self.contextHandler.addBundleFile([fileName])
			elif extension in self.contextHandler.validMeshExtension():
				self.contextHandler.addMeshFile([fileName])
			elif extension in self.contextHandler.validMRIExtension():
				self.contextHandler.addMRIFile([fileName])
			# elif extension == "bundlesdata":
			# 	self.contextHandler.addBundleFile([fileName[:-4]])
			else:
				print('File extension not sopported: ', fileName)


	def closeEvent(self, event):
		print('Cleaning...')
		self.contextHandler.cleanOpenGL()

		super().closeEvent(event)


	def keyPressEvent(self, event):
		# print('Key press: {0}'.format(event.key()))
		pass
		# self.contextHandler.setWindowState(QtCore.Qt.WindowFullScreen)
		# self.ui.showFullScreen()
		# if self.tFullScreen and event.key() == QtCore.Qt.Key_Escape:
		# 	event.accept()
		# 	self.toggleFullScreen()
		# else:
		# 	event.ignore()


	@QtCore.pyqtSlot()
	def unCheckActionVisualizationObjects(self):
		self.ui.actionVisualizationObjects.setChecked(False)


	@QtCore.pyqtSlot()
	def openBundleFile(self):
		''' Function that gets called when pressing the open bundle file action
		'''

		# Gets direction of bundle files
		formats = "" # *.bundles *.trk *.tck
		for f in self.contextHandler.validBundleExtension():
			formats += "*."+f+" "
		bundleFiles, fileType = QtWidgets.QFileDialog.getOpenFileNames(self, "Select Bundle", "", "Bundles ("+formats+")")

		if len(bundleFiles) == 0:
			return

		self.contextHandler.addBundleFile(bundleFiles)


	@QtCore.pyqtSlot()
	def openMeshFile(self):
		''' Function that gets called when pressing the open mesh file action
		'''

		# Gets direction of bundle files
		formats = ""
		for f in self.contextHandler.validMeshExtension():
			formats += "*."+f+" "
		meshFiles, fileType = QtWidgets.QFileDialog.getOpenFileNames(self, "Select Mesh", "", "Mesh ("+formats+")")

		if len(meshFiles) == 0:
			return

		self.contextHandler.addMeshFile(meshFiles)


	@QtCore.pyqtSlot()
	def openMRIFile(self):
		''' Function that gets called when pressing the open MRI file action
		'''

		# Gets direction of bundle files
		formats = ""
		for f in self.contextHandler.validMRIExtension():
			formats += "*."+f+" "
		mriFiles, fileType = QtWidgets.QFileDialog.getOpenFileNames(self, "Select MRI", "", "MRI ("+formats+")")

		if len(mriFiles) == 0:
			return

		self.contextHandler.addMRIFile(mriFiles)


	@QtCore.pyqtSlot()
	def createROI(self):
		'''
		'''

		self.contextHandler.addROI()


	def testSlot(self):
		print('testSlot!')


from collections import deque
from enum import Enum
import numpy as np


class TokenType(Enum):
	NUMBER = 0
	OPERATOR = 1
	LEFTBRACKET = 2
	RIGHTBRACKET = 3


class Associative(Enum):
	LEFT = 0
	RIGHT = 1


class Token:
	value = None
	tokenType = None

	def __init__(self, value=None):
		if value is None:
			return

		self.value = value

		if isinstance(value, str) and len(value) == 1:
			if value == '(':
				self.tokenType = TokenType.LEFTBRACKET
			elif value == ')':
				self.tokenType = TokenType.RIGHTBRACKET
			else:
				self.tokenType = TokenType.OPERATOR
		else:
			self.tokenType = TokenType.NUMBER
			

class LogicalMath:

	MathTable = {'&' : {'func' : np.logical_and, 'nOfOperand' : 2},
				'|' : {'func' : np.logical_or, 'nOfOperand' : 2},
				'!' : {'func' : np.logical_not, 'nOfOperand' : 1},
				'^' : {'func' : np.logical_xor, 'nOfOperand' : 2}}

	PrecedenceTable = {	'|' : 0,
						'^' : 0,
						'&' : 1,
						'!' : 2}

	AssociativeTable = {'&' : Associative.LEFT,
						'|' : Associative.LEFT,
						'!' : Associative.LEFT,
						'^' : Associative.LEFT}


class Math:
	MathTable = {	'^' : {'func' : lambda a,b : pow(a,b), 'nOfOperand' : 2},
					'*' : {'func' : lambda a,b : a*b, 'nOfOperand' : 2},
					'/' : {'func' : lambda a,b : a/b, 'nOfOperand' : 2},
					'+' : {'func' : lambda a,b : a+b, 'nOfOperand' : 2},
					'-' : {'func' : lambda a,b : a-b, 'nOfOperand' : 2}}

	PrecedenceTable = {	'^' : 4,
						'*' : 3,
						'/' : 3,
						'+' : 2,
						'-' : 2}

	AssociativeTable = {'(', Associative.LEFT,
						'^', Associative.RIGHT,
						'*', Associative.LEFT,
						'/', Associative.LEFT,
						'+', Associative.LEFT,
						'-', Associative.LEFT}


def infix2postfix(tokens, precedenceTable):
	if not isinstance(tokens, deque):
		print('shunting_yard: infix2postfix input is not a deque... casting into one.')
		tokens = deque(tokens)

	operatorStack = []
	outputqueue = deque()

	while len(tokens):
		token = tokens.popleft()

		if token.tokenType == TokenType.NUMBER:
			outputqueue.append(token)

		# elif token in functionTable:
			# operatorStack.append(token)

		elif token.tokenType == TokenType.OPERATOR:
			while ( len(operatorStack) and 
				operatorStack[-1].tokenType != TokenType.LEFTBRACKET and 
				precedenceTable[token.value] < precedenceTable[operatorStack[-1].value]):

				outputqueue.append(operatorStack.pop())
			operatorStack.append(token)

		elif token.tokenType == TokenType.LEFTBRACKET:
			operatorStack.append(token)

		elif token.tokenType == TokenType.RIGHTBRACKET:
			while len(operatorStack) and operatorStack[-1].tokenType != TokenType.LEFTBRACKET:
				outputqueue.append(operatorStack.pop())

			if not len(operatorStack):
				return -1

			operatorStack.pop()

	while len(operatorStack):
		outputqueue.append(operatorStack.pop())
		if outputqueue[-1].tokenType == TokenType.LEFTBRACKET:
			return -1

	return outputqueue


def postfixEvaluator(tokens, MathTable):
	if not isinstance(tokens, deque):
		print('shunting_yard: postfixEvaluator input is not a deque... casting into one.')
		tokens = deque(tokens)

	stack = []

	while len(tokens):
		token = tokens.popleft()

		if token.tokenType == TokenType.NUMBER:
			stack.append(token)

		else:
			right_operand = stack.pop()

			if MathTable[token.value]['nOfOperand'] == 1:
				stack.append(Token(MathTable[token.value]['func'](right_operand.value)))
			else:
				left_operand = stack.pop()
				stack.append(Token(MathTable[token.value]['func'](left_operand.value, right_operand.value)))

	if len(stack) == 1:
		return stack.pop()

	else:
		raise ValueError


def shuntingYard(infix, operation='Logical'):
	if operation == 'Logical':
		table = LogicalMath
	elif operation == 'Math':
		table = Math
	else:
		raise ValueError('Not recogniced.')

	infixTokens = deque([Token(i) for i in infix])
	postfixTokens = infix2postfix(infixTokens, table.PrecedenceTable)

	if postfixTokens == -1:
		raise ValueError('shunting_yard: Bracket not paired.')

	# print([i.value for i in postfixTokens])

	evaluation = postfixEvaluator(postfixTokens, table.MathTable)

	return evaluation.value


"""
Phybers is a library for Python that integrates multiple tractography and neural-fibers related algorithms. It includes the following sub-packages:
    phybers.clustering "Clustering"
    phybers.segment "Segment"
    phybers.utils "Utilities"
    phybers.fibervis "Fibervis"
The functions of this package work mostly with file paths, both for input and output.
"""

__all__ = ['clustering', 'fibervis', 'segment', 'utils']


# -*- coding: utf-8 -*-
import sys, os
import numpy as N
from time import time
from scipy import sparse



#class with Average Link Analisys vars
class WeightedGraph():


  def __init__( self, V, edges, weights ):

    self.V = V
    self.E = len( edges )
    self.edges = edges
    self.weights = weights


def createfffGraphFromFile( graphfile ):

  #read graph file and create fff graph
  inFile = open( graphfile, 'r' )
  line = inFile.readline()
  l = line.split()
  V = int( l[ 0 ] )
  E = int( l[ 1 ] )
  edges = N.zeros( [ E, 2 ], 'i' )
  weights = N.zeros( [ E, 1 ], 'f' )
  print ('edges: ', E, ' vertices: ', V)

  count = 0
  while True:

    line = inFile.readline()
    if line == '': break
    l = line.split()
    edges[ count, 0 ] = int( l[ 0 ] )
    edges[ count, 1 ] = int( l[ 1 ] )
    weights[ count ] = float( l[ 2 ] )
    count += 1
  inFile.close()
  G = WeightedGraph( V, edges, weights )
  return G


#class with Average Link Analisys vars
class CenterWForestVars():


  def __init__( self, wforestFile ):

    f = open( wforestFile, 'r' )

    #number of Wforest nodes
    self.V = int( f.readline() )
    #list of node father
    self.fa = []
    #list of node child indexes (two)
    self.ch = []

    for v in range( self.V ):

      line = f.readline()
      line = line.split()
      self.fa.append( int( line[ 0 ] ) )
      child = [ int( line[ 1 ] ), int( line[ 2 ] ) ]
      self.ch.append( child )

    f.close()
    #number of white matter parcels
    self.Vg = 0
    #points not included in the Weighted Forest (without connectivity)
    self.singles = []
    #number of parcels included in the Weighted Forest (clusterized)
    self.VV = 0
    #number of trees
    self.top = []
    #final top nodes of partition
    self.finalcl = []
    #singles additional nodes after partition
    self.singles2 = []
    #final clusters
    self.clusters = []

  def get_descendents( self, v ):

    v = int( v )
    if ( v >= 0 and v <= self.V - 1 ):

      if ( self.ch[ v ][ 0 ] == -1 and self.ch[ v ][ 1 ] == -1 ):

        return [ v ]

      else:

        descendent = [ v ]

        for c in self.ch[ v ]:

          temp = self.get_descendents( c )

          for q in temp:

            descendent.append( q )

      return descendent


##for clusterizeCentersMaxdist.py
def createSparceScipyMatrixFromfffGraph( graphfile ):

  G = createfffGraphFromFile( graphfile )
  V = G.V
  edges = G.edges
  weights = G.weights
  mat = sparse.lil_matrix( ( V, V ) )
  for ind in range( len( edges ) ):

    ed = edges[ ind ]
    i = ed[ 0 ]
    j = ed[ 1 ]
    mat[ i,j ] = weights[ ind ].item()
  return mat


#calculate max and mean distances within cluster nodes
#(clusters are connected components of the tree and not singles)
def calculate_top_dist( WFV, data, top, maxdist ):

  #number of default clusters
  #(connected components of the tree and not singles)
  defcl = len( top )
  dist_data  = N.zeros( defcl ) # max
  for i in range( defcl ):

    t = top[ i ]
    maxd = calculate_node_dist( WFV, data, t, maxdist )
    dist_data[ i ] = maxd
  return dist_data

#calculate max and mean distances within cluster nodes
#(clusters are connected components of the tree and not singles)
def calculate_top_dist_nomatrix( WFV, sparse_mat, top, minaff ):

  #number of default clusters
  #(connected components of the tree and not singles)
  defcl = len( top )
  dist_data  = N.zeros( defcl ) # max
  for i in range( defcl ):

    t = top[ i ]
    mind = calculate_node_dist_nomatrix( WFV, sparse_mat, t, minaff )
    dist_data[ i ] = mind
  return dist_data

#calculate max and mean distances within cluster nodes
#(clusters are connected components of the tree and not singles)
def calculate_node_dist( WFV, data, node, maxdist ):

  maxd = 0.0
  if ( WFV.ch[ node ][ 0 ] != -1 and WFV.ch[ node ][ 1 ] != -1 ):

    #get all cluster nodes
    s = WFV.get_descendents( node )
    s.remove( node )
    ss = N.array( s )
    #identify leaves
    ind_s = N.where( ss < WFV.Vg )[ 0 ]
    s = ss[ ind_s ]
    for s1 in range( len( s ) ):

      a = s[ s1 ]
      for s2 in range( s1 + 1, len( s ) ):

        b = s[ s2 ]
        val = data[ a ][ b ]
        if val > maxd:

          maxd = val
        if maxd > maxdist:

          return val
  return maxd

#(clusters are connected components of the tree and not singles)
def calculate_node_dist_nomatrix( WFV, sparse_mat, node, minaff ):

  #node = WFV.top(1)
  mind = 100.0
  if ( WFV.ch[ node ][ 0 ] != -1 and WFV.ch[ node ][ 1 ] != -1 ):

    #get all cluster nodes
    s = WFV.get_descendents( node )
    s.remove( node )
    ss = N.array( s )
    #identify leaves
    ind_s = N.where( ss < WFV.Vg )[ 0 ]
    s = ss[ ind_s ]
    for s1 in range( len( s ) ):

      a = s[ s1 ]
      #get all edges indexes who have the node
      for s2 in range( s1 + 1, len( s ) ):

        b = s[ s2 ]
        val = sparse_mat[ a, b ]
        if val < mind:

          mind = val
        if mind <= minaff:

          return mind
  return mind


def create_partition_from_finalcl( WFV, finalcl ):

  clusters = []
  for f in finalcl:

    if ( WFV.ch[ f ][ 0 ] != -1 and WFV.ch[ f ][ 1 ] != -1 ): #not single

      s = WFV.get_descendents( f )
      s.remove( f )
      ss = N.array( s )
      #identify leaves
      ind_s = N.where( ss < WFV.Vg )[ 0 ]
      s = ss[ ind_s ]
      nodelist = list( s )
    else:

      nodelist = [ f ]
    clusters.append( nodelist )

  WFV.clusters = clusters

def wforest_partition_maxdist_from_graph( wforestfile,
                                       maxdist,
                                       add_singles = False,
                                       graphfile = '',
                                       var: float = 3600 ):

  WFV = CenterWForestVars( wforestfile )
  nomatrix=True
  #number of vertices of the graph (fiber centers)
  Vg = 0
  while ( Vg < len( WFV.ch ) and \
                         WFV.ch[ Vg ][ 0 ] == -1 and WFV.ch[ Vg ][ 1 ] == -1  ):

    Vg = Vg + 1

  singles = []
  for v in range( Vg ):

    if WFV.fa[ v ] == v:
      singles.append( v )
  #
  #number of centers included in the Weighted Forest (clusterized)
  VV = Vg - len( singles )
  #
  #number of trees
  top = []
  for v in range( Vg, WFV.V ):

    if WFV.fa[ v ] == v:

      top.append( v )
  #store WFV data
  WFV.Vg = Vg
  WFV.singles = singles
  WFV.VV = VV
  WFV.top = top

  #calculate maximum values of distances to evaluate clusters
  #print("maxdist type: ")
  print(maxdist)
  print(type(maxdist))
  max_cldist = float(maxdist) * 1.0         #maximum distance within a cluster
  max_meancldist = float(maxdist) * 1.0      #mean distance within a cluster
  #minimum affinity within a cluster
  print(type(-max_cldist))
  print(type(max_cldist))
  print(type(var))
  min_claff = N.exp( -max_cldist * max_cldist / float(var) );
  #mean affinity within a cluster
  min_meanclaff = N.exp( -max_meancldist * max_meancldist / float(var) );

  #---------------------------------------------
  #information of distance within clusters
  if nomatrix:

    #convert fffgraph to scipy sparse matrix
    sparse_mat = createSparceScipyMatrixFromfffGraph( graphfile )
    dist_data = calculate_top_dist_nomatrix( WFV, sparse_mat, top, min_claff )
  #else:

    #mat = aims.read( all_centers_distmat )
    #data = mat.arraydata()[ 0 ][ 0 ]
    #calculate distance data
    #dist_data = calculate_top_dist( WFV, data, top, max_cldist )

  count_divide = 0 #tmp
  singles2b = []
  #review only non-tight clusters
  defcl = len( top )
  finalcl = []
  top2 = []

  for i in range( defcl ):

    maxd = dist_data[ i ]
    divide = False
    if nomatrix:

      if maxd <= min_claff: #maxd in the minimum affinity

        divide = True

    if divide:

       #add top chidren to divide queue
      count_divide += 1 #tmp
      t = top[ i ]
      children = WFV.ch[ t ]
      if ( children[ 0 ] != -1 or children[ 1 ] != -1 ):

        for chi in children:

          if chi >= Vg:

            top2.append( chi )
          else:

            singles2b.append( chi )
    else:

      finalcl.append( top[ i ] )

  #analisys queue
  queue = []
  #add all trees in queue
  for i in range( len( top2 ) ):

    queue.append( top2[ i ] )

  singles2 = []
  #perform analisys
  while len( queue ) > 0:

    ind = queue[ 0 ]
    queue.remove( ind )
    if nomatrix:

      maxd = calculate_node_dist_nomatrix( WFV,
                                           sparse_mat,
                                           ind,
                                           min_claff )


    children = WFV.ch[ ind ]
    #if maxd > max_cldist and meand > max_meancldist:
    divide = False
    if nomatrix:

      if maxd <= min_claff: #maxd in the minimum affinity

        divide = True

    if divide:

      count_divide += 1 #tmp
      #must divide
      if ( WFV.ch[ Vg ][ 0 ] != -1 and WFV.ch[ Vg ][ 1 ] != -1 ):

        for chi in children:

          if chi >= Vg:

            queue.append( chi )
          else:

            singles2b.append( chi )
      else:

        singles2.append( ind )
    else:

      if ( WFV.ch[ Vg ][ 0 ] != -1 and WFV.ch[ Vg ][ 1 ] != -1 ):

        finalcl.append( ind )
      else:

        singles2.append( ind )

  WFV.singles2 = singles2 + singles2b
  WFV.finalcl_singles = singles + singles2 + singles2b
  WFV.finalcl_notsingles = finalcl
  WFV.finalcl = finalcl + WFV.finalcl_singles
  print ('total singles num = ', len( WFV.finalcl_singles ))
  print ('not single clusters = ', len( WFV.finalcl_notsingles ))
  print ('total clusters = ', len( WFV.finalcl ))
  #tmp
  print ('count divide = ', count_divide)

  if add_singles == True:

    create_partition_from_finalcl( WFV, WFV.finalcl )
  else:

    create_partition_from_finalcl( WFV, WFV.finalcl_notsingles )
  return WFV



from .Quaternion import Quaternion
import numpy as np

def null(w=4, h=4):
	return np.matrix([[ 0 for i in range(w)] for j in range(h)], dtype = "float32")


def identity(w=4, h=4):
	return np.matrix([[1 if i==j else 0 for i in range(w)] for j in range(h)], dtype = "float32")


def normalize(v):
	module = np.sqrt(np.multiply(v[0:3],v[0:3]).sum())

	v /= module

	return v


def translateMatrix(v):
	return np.matrix([	[1,	0,	0,	v[0]],
						[0,	1,	0,	v[1]],
						[0,	0,	1,	v[2]],
						[0,	0,	0,	1]], dtype=np.float32)


def scaleMatrix(v):
	if hasattr(v, '__iter__'):
		return np.matrix([	[v[0],	0,		0,		0],
							[0,		v[1],	0,		0],
							[0,		0,		v[2],	0],
							[0,		0,		0,		1]], dtype=np.float32)
	else:
		return np.matrix([	[v,		0,		0,		0],
							[0,		v,		0,		0],
							[0,		0,		v,		0],
							[0,		0,		0,		1]], dtype=np.float32)


def rotateMatrix(alfa, axis):
	return Quaternion.fromAngleAxis(alfa*np.pi/180, axis).rotation4Matrix()


def ortho(left, right, bottom, top, zNear, zFar):
	Result = identity()
	Result[0,0] = 2.0 / (right - left);
	Result[1,1] = 2.0 / (top - bottom);
	Result[2,2] = -2.0 / (zFar - zNear);
	Result[3,0] = -(right + left) / (right - left);
	Result[3,1] = -(top + bottom) / (top - bottom);
	Result[3,2] = -(zFar + zNear) / (zFar - zNear);

	return Result;


def frustum(left, right, bottom, top, nearVal, farVal):
	Result = null()
	Result[0,0] = (2.0 * nearVal) / (right - left)
	Result[1,1] = (2.0 * nearVal) / (top - bottom)
	Result[2,0] = (right + left) / (right - left)
	Result[2,1] = (top + bottom) / (top - bottom)
	Result[2,2] = -(farVal + nearVal) / (farVal - nearVal)
	Result[2,3] = -1.0
	Result[3,2] = -(2.0 * farVal * nearVal) / (farVal - nearVal)

	return Result


def perspective(fovy, aspect, zNear, zFar):
	rad = np.radians(fovy)

	tanHalfFovy = np.tan(rad/2.0)

	Result = null()
	Result[0,0] = 1.0/(aspect*tanHalfFovy)
	Result[1,1] = 1.0/tanHalfFovy
	Result[2,2] = -(zFar + zNear)/(zFar - zNear)
	Result[2,3] = -(2*zFar*zNear)/(zFar - zNear)
	Result[3,2] = -1.0


	return Result


def lookAt(eye, center, up):
	# f = normalize(eye-center)
	# u = normalize(up)
	# s = normalize(np.cross(u, f))
	# u = np.cross(f,s)

	# rotation = identity()

	# rotation[0,0] = s[0];
	# rotation[0,1] = s[1];
	# rotation[0,2] = s[2];

	# rotation[1,0] = u[0];
	# rotation[1,1] = u[1];
	# rotation[1,2] = u[2];

	# rotation[2,0] = f[0];
	# rotation[2,1] = f[1];
	# rotation[2,2] = f[2];

	# rotation[3,0] = -eye[0]
	# rotation[3,1] = -eye[1]
	# rotation[3,2] = -eye[2]

	f = normalize(center-eye)
	u = normalize(up)
	s = normalize(np.cross(f, u))
	u = np.cross(s, f)

	Result = identity()

	Result[0,0] = s[0];
	Result[1,0] = s[1];
	Result[2,0] = s[2];
	Result[0,1] = u[0];
	Result[1,1] = u[1];
	Result[2,1] = u[2];
	Result[0,2] = -f[0];
	Result[1,2] = -f[1];
	Result[2,2] = -f[2];
	Result[3,0] = -np.dot(s, eye)
	Result[3,1] = -np.dot(u, eye)
	Result[3,2] = np.dot(f, eye)

	return Result

	# print(rotation)

	# return rotation.T


def inverseTranspose(m):
	SubFactor00 = m[2,2] * m[3,3] - m[3,2] * m[2,3]
	SubFactor01 = m[2,1] * m[3,3] - m[3,1] * m[2,3]
	SubFactor02 = m[2,1] * m[3,2] - m[3,1] * m[2,2]
	SubFactor03 = m[2,0] * m[3,3] - m[3,0] * m[2,3]
	SubFactor04 = m[2,0] * m[3,2] - m[3,0] * m[2,2]
	SubFactor05 = m[2,0] * m[3,1] - m[3,0] * m[2,1]
	SubFactor06 = m[1,2] * m[3,3] - m[3,2] * m[1,3]
	SubFactor07 = m[1,1] * m[3,3] - m[3,1] * m[1,3]
	SubFactor08 = m[1,1] * m[3,2] - m[3,1] * m[1,2]
	SubFactor09 = m[1,0] * m[3,3] - m[3,0] * m[1,3]
	SubFactor10 = m[1,0] * m[3,2] - m[3,0] * m[1,2]
	SubFactor11 = m[1,1] * m[3,3] - m[3,1] * m[1,3]
	SubFactor12 = m[1,0] * m[3,1] - m[3,0] * m[1,1]
	SubFactor13 = m[1,2] * m[2,3] - m[2,2] * m[1,3]
	SubFactor14 = m[1,1] * m[2,3] - m[2,1] * m[1,3]
	SubFactor15 = m[1,1] * m[2,2] - m[2,1] * m[1,2]
	SubFactor16 = m[1,0] * m[2,3] - m[2,0] * m[1,3]
	SubFactor17 = m[1,0] * m[2,2] - m[2,0] * m[1,2]
	SubFactor18 = m[1,0] * m[2,1] - m[2,0] * m[1,1]

	Inverse = null()
	Inverse[0,0] = + (m[1,1] * SubFactor00 - m[1,2] * SubFactor01 + m[1,3] * SubFactor02)
	Inverse[0,1] = - (m[1,0] * SubFactor00 - m[1,2] * SubFactor03 + m[1,3] * SubFactor04)
	Inverse[0,2] = + (m[1,0] * SubFactor01 - m[1,1] * SubFactor03 + m[1,3] * SubFactor05)
	Inverse[0,3] = - (m[1,0] * SubFactor02 - m[1,1] * SubFactor04 + m[1,2] * SubFactor05)

	Inverse[1,0] = - (m[0,1] * SubFactor00 - m[0,2] * SubFactor01 + m[0,3] * SubFactor02)
	Inverse[1,1] = + (m[0,0] * SubFactor00 - m[0,2] * SubFactor03 + m[0,3] * SubFactor04)
	Inverse[1,2] = - (m[0,0] * SubFactor01 - m[0,1] * SubFactor03 + m[0,3] * SubFactor05)
	Inverse[1,3] = + (m[0,0] * SubFactor02 - m[0,1] * SubFactor04 + m[0,2] * SubFactor05)

	Inverse[2,0] = + (m[0,1] * SubFactor06 - m[0,2] * SubFactor07 + m[0,3] * SubFactor08)
	Inverse[2,1] = - (m[0,0] * SubFactor06 - m[0,2] * SubFactor09 + m[0,3] * SubFactor10)
	Inverse[2,2] = + (m[0,0] * SubFactor11 - m[0,1] * SubFactor09 + m[0,3] * SubFactor12)
	Inverse[2,3] = - (m[0,0] * SubFactor08 - m[0,1] * SubFactor10 + m[0,2] * SubFactor12)

	Inverse[3,0] = - (m[0,1] * SubFactor13 - m[0,2] * SubFactor14 + m[0,3] * SubFactor15)
	Inverse[3,1] = + (m[0,0] * SubFactor13 - m[0,2] * SubFactor16 + m[0,3] * SubFactor17)
	Inverse[3,2] = - (m[0,0] * SubFactor14 - m[0,1] * SubFactor16 + m[0,3] * SubFactor18)
	Inverse[3,3] = + (m[0,0] * SubFactor15 - m[0,1] * SubFactor17 + m[0,2] * SubFactor18)

	Determinant = m[0,0] * Inverse[0,0] + m[0,1] * Inverse[0,1] + m[0,2] * Inverse[0,2] + m[0,3] * Inverse[0,3]

	Inverse /= Determinant

	return Inverse


from PyQt5 import QtGui, QtWidgets, uic, QtCore
from importlib_resources import files

_aipsd_ui = files('phybers.fibervis.ui.Segmentations').joinpath(
	'InPlaceSegmentationDialog.ui')

class InPlaceSegmentationDialog(QtWidgets.QDialog):
	updateObject = QtCore.pyqtSignal(dict)


	def __init__(self, segmentation, parent):
		super().__init__(parent)
		self.ui = uic.loadUi(str(_aipsd_ui), self)

		self.segmentationRef = segmentation

		self.bundleNames = None
		self.bundleSize = None
		self.states = []
		self.setBundles(self.segmentationRef.bundlesName, self.segmentationRef.bundlesStart)

		self.ui.tableWidget.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
		self.ui.tableWidget.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
		self.ui.tableWidget.setColumnWidth(0, 150)
		self.ui.tableWidget.horizontalHeader().setStretchLastSection(True)

		self.ui.tableWidget.cellClicked.connect(self.toggleCheck)

		self.ui.checkVisualizeAll.clicked.connect(self.visualizeAll)

		self.ui.percentSlider.valueChanged.connect(self.checkRenderState)

		self.ui.filterLineEdit.textChanged.connect(self.filter)

		self.ui.alphaSlider.valueChanged.connect(self.alphaChanged)

		self.ui.renderButton.clicked.connect(self.callUpdate)
		self.ui.exportBundleButton.clicked.connect(self.exportBundle)

		self.show()


	@QtCore.pyqtSlot(int)
	def toggleCheck(self, row):
		tmpItem = self.ui.tableWidget.item(row,0)
		tmpItem.setCheckState(not tmpItem.checkState())


	@QtCore.pyqtSlot()
	def visualizeAll(self):
		self.ui.tableWidget.cellChanged.disconnect(self.toggleStates)

		visualize = self.ui.checkVisualizeAll.isChecked()

		for i in range(self.ui.tableWidget.rowCount()):
			item = self.ui.tableWidget.item(i,0)
			item.setCheckState(visualize)
			self.states[self.bundleNames.index(item.text())] = visualize

		if self.ui.checkRender.isChecked():
			self.updateObject.emit(self.returnDict())

		self.ui.tableWidget.cellChanged.connect(self.toggleStates)


	@QtCore.pyqtSlot()
	def callUpdate(self):
		self.updateObject.emit(self.returnDict())


	@QtCore.pyqtSlot()
	def checkRenderState(self):
		if self.ui.checkRender.isChecked():
			self.updateObject.emit(self.returnDict())


	# TMP!!!!!!
	@QtCore.pyqtSlot(str)
	def filter(self, text):
		self.ui.tableWidget.cellChanged.disconnect(self.toggleStates)

		self.ui.tableWidget.clearContents()
		self.ui.tableWidget.setRowCount(0)

		row = 0
		for i in range(len(self.bundleNames)):
			if self.bundleNames[i].find(text) != -1:
				self.ui.tableWidget.insertRow(row)
				item = QtWidgets.QTableWidgetItem(self.bundleNames[i])
				item.setCheckState(self.states[i])
				self.ui.tableWidget.setItem(row, 0, item)

				self.ui.tableWidget.setItem(row, 1, QtWidgets.QTableWidgetItem(str(self.bundleSize[i])))
				self.ui.tableWidget.item(row,1).setTextAlignment(QtCore.Qt.AlignRight)
				row = row + 1

		checkStateN = 0

		for i in range(self.ui.tableWidget.rowCount()):
			if self.ui.tableWidget.item(i,0).checkState():
				checkStateN += 1

		if checkStateN == self.ui.tableWidget.rowCount():
			self.ui.checkVisualizeAll.setChecked(True)
		elif checkStateN == 0:
			self.ui.checkVisualizeAll.setChecked(False)

		self.ui.tableWidget.cellChanged.connect(self.toggleStates)


	@QtCore.pyqtSlot(int)
	def toggleStates(self, row):
		item = self.ui.tableWidget.item(row,0)
		tmp = self.bundleNames.index(item.text())
		self.states[tmp] = bool(item.checkState())

		tmp = 0

		for i in range(self.ui.tableWidget.rowCount()):
			tmp += int(self.ui.tableWidget.item(i,0).checkState())

		if tmp == self.ui.tableWidget.rowCount():
			self.ui.checkVisualizeAll.setChecked(True)
		else:
			self.ui.checkVisualizeAll.setChecked(False)

		if self.ui.checkRender.isChecked():
			self.updateObject.emit(self.returnDict())


	@QtCore.pyqtSlot()
	def exportBundle(self):

		# Gets direction of bundle files
		outFile, fileType = QtWidgets.QFileDialog.getSaveFileName(self, "Save Bundle file", "", "Bundle (*.bundles)")#;;Trk file (*.trk)")

		options = {	'reference' : self.segmentationRef,
					'attributes' : {'exportFile' : [outFile]}}

		self.updateObject.emit(options)


	@QtCore.pyqtSlot()
	def alphaChanged(self):
		alpha = self.alphaSlider.value()/(self.alphaSlider.maximum()+1)

		options = {	'reference' : self.segmentationRef,
					'attributes': {'setAlpha' : [alpha]}}

		self.updateObject.emit(options)


	def setBundles(self, bundlenames, bundlesize):
		self.bundleNames = bundlenames
		self.bundleSize = bundlesize[1:]-bundlesize[:-1]

		# self.ui.tableWidget.cellChanged.disconnect(self.toggleStates)

		self.ui.tableWidget.clearContents()

		for i in range(len(self.bundleNames)):
			self.states.append(True)

			self.ui.tableWidget.insertRow(i)
			item = QtWidgets.QTableWidgetItem(self.bundleNames[i])
			item.setCheckState(True)
			self.ui.tableWidget.setItem(i,0,item)

			self.ui.tableWidget.setItem(i,1,QtWidgets.QTableWidgetItem(str(self.bundleSize[i])))
			self.ui.tableWidget.item(i,1).setTextAlignment(QtCore.Qt.AlignRight)

		self.ui.checkVisualizeAll.setChecked(True)

		self.ui.tableWidget.cellChanged.connect(self.toggleStates)

	def returnDict(self):
		return {'reference' : self.segmentationRef,
				'attributes' : {'setPercentage' : [self.ui.percentSlider.value()],
								'setStates' : [self.states]},
				'segmentSubject' : True}


	def __del__(self):
		print('Must clean befor leaving... INPLACESEGMENTATIONDIALOG', type(self))
		segmentationRef = None
		# super.__del__()


	def update(self):
		pass


"""Segmentation Module
"""

import os
import sys
import numpy as np
from pathlib import Path
from .fiberseg.c_wrappers import segment
from ..utils import sampling, read_bundle, write_bundle


def fiberseg(file_in: str, id_subj: str, atlas_dir: str, atlas_info: str, dir_out: str) -> None:
    """
    White matter fiber bundle segmentation algorithm based on a multi-subject atlas.
    
    Parameters
    ----------
    file_in : str
        Tractography data file in *'.bundles'* format.
    id_subj : str
        Subject identification number, used to label the results.
    atlas_dir : str
        Bundle atlas, with bundles in separated files, sampled at 21 equidistant points. The bundle atlases provided are in same folders.
    atlas_info : str
        Text file that stores information about the brain fiber atlas, such as the fascicle name, segmentation threshold, and fascicle size.
        Note that the segmentation threshold can be adjusted depending on the database to be used.
    dir_out : str
        Directory name to store all the results generated by the algorithm.

    Return
    ------
    None

    Notes
    -----
    This function generates the following files in the specified directory:    
    
    Segmented fibers : bundles files
        Directory contains all atlas fascicles extracted from the subject, saved as separate files in the '.bundles' format. 
        Each file's name is composed of the atlas label followed by the subject's ID
    Centroids : bundles file
        Directory that contains the centroid for each fascicle segmented in same *'.bundles'* files. 
    Index of fibers per fasciculus : text file
        Text file containing the indexes of the fibers that were segmented by each fascicle of the atlas.
        
    Examples
    --------
    To use the `fiberseg()` function, you need to download the tractography data from the `link fiberseg  <https://www.dropbox.com/sh/tj67742sxvmqfg1/AAD_J41rw4E70OVFayAy18T6a?dl=1>`_ .
    Then, you can run the following code in a Python terminal:

    >>> from phybers.segment import fiberseg
    >>> file_in = 'path_to_tractography_data.bundles'
    >>> id_subj = '01'
    >>> atlas_dir = 'path_to_atlas/bundles'
    >>> atlas_info = 'path_to_atlas/atlas_info.txt'
    >>> dir_out = 'path_to_output_directory'
    >>> fiberseg(file_in, id_subj, atlas_dir, atlas_info, dir_out)

    Note: Make sure to replace 'path_to_tractography_data', 'path_to_atlas', and 'path_to_output_directory' with the actual paths to your data and directories. 
    
    """


    id_seg_result= os.path.join(dir_out,'idx_bundles')
    os.makedirs(id_seg_result, exist_ok=True)

    has21points = True
    data=read_bundle(file_in)

    for i in range(len(data)-1):
        if len(data[i]) != len(data[i+1]):
            has21points = False
            break

    final_bundles21p_dir = os.path.join(dir_out, 'FinalBundles21p')
    os.makedirs(final_bundles21p_dir, exist_ok=True)

    if not has21points:

        final_bundles_dir = os.path.join(dir_out, 'FinalBundles')
        os.makedirs(final_bundles_dir, exist_ok=True)

        outfile_dir= os.path.join(dir_out, 'outfile')
        os.makedirs(outfile_dir, exist_ok=True)

        fibers21p = os.path.join(outfile_dir,'fiberorig_21p.bundles')

        sampling(file_in, fibers21p, 21)

        segment(21, fibers21p, id_subj, atlas_dir, atlas_info,
                final_bundles21p_dir, id_seg_result)

        for i in os.listdir(id_seg_result):
            index=[]
            with open (os.path.join(id_seg_result,i)) as file:
                for line in file:
                    index.append(int(float(line.strip())))
            bun = np.array(data, dtype='object')[index]
            write_bundle(os.path.join(final_bundles_dir,id_subj+'_to_'+i[:-4]+'.bundles'), bun)

    else:
        segment(21, file_in, id_subj, atlas_dir, atlas_info,
                final_bundles21p_dir, id_seg_result)



import numpy as np
from .SegmentationHandler import SegmentationHandler
from ..BoundingBox import BoundingBox
from ..Tools.visualizationEnums import SegmentationTypes
from ...FiberVis_core import inPlaceSegmentationMethod

class InPlaceSegmentation(SegmentationHandler):
	def __init__(self, bundle, shaderDict):
		super().__init__(bundle, shaderDict)

		self.segmentationIdentifier = SegmentationTypes.InPlace

		self.bundleStates = np.ones(len(self.bundlesName), dtype=np.int32) # int8 = c boolean
		self.percentage = 100
		self.fileName = 'In place segmentation' # temporal

		self.alpha = 0.8

		# Creates VBOs & an EBO, then populates them with the point, normal and color data. The elements data goes into the EBO
		self.configFiberValidator()

		self._loadBuffers()
		self.buildVertex2Fiber()
		self.vboAndLinkVertex2Fiber()

		self.boundingbox = BoundingBox(shaderDict, self, bundle.boundingbox.dims, bundle.boundingbox.center)


	def segmentMethod(self):
		self.fiberValidator[:self.curvescount] = 0

		inPlaceSegmentationMethod(
			self.bundleStates.size,
			self.percentage,
			self.bundlesStart,
			self.bundleStates,
			self.fiberValidator)

		# cfuncs.inPlaceSegmentationMethod(
		# 	self.bundleStates.size,
		# 	self.percentage,
		# 	self.bundlesStart.ctypes.data,
		# 	self.bundleStates.ctypes.data,
		# 	self.fiberValidator.ctypes.data)


	def setPercentage(self, newPercentage):
		self.percentage = newPercentage


	def setStates(self, newStates):
		self.bundleStates = np.array(newStates, dtype=np.int32)


import numpy as np
from ...FiberVis_core import ROIsSegmentationQueryOctree, ROIsSegmentationPopulateAndDefragmentPool


class OctreePointBased:
	OT_NODE_CAPACITY = 256

	def __init__(self, points, fiberSizes):
		print('Building Octree...')
		self.vertex = points
		self.treeCount = 0
		self.root = 0
		self.pool = None
		self.vertexIdxPool = None

		x = self.vertex[0::3]
		y = self.vertex[1::3]
		z = self.vertex[2::3]

		xmin, xmax = x.min(),x.max()
		ymin, ymax = y.min(),y.max()
		zmin, zmax = z.min(),z.max()

		radius = np.array([(xmax-xmin)/2, (ymax-ymin)/2, (zmax-zmin)/2], dtype=np.float32) * 1.005
		center = radius + np.array([xmin, ymin, zmin], dtype=np.float32)

		self.CPopulateAndDefragment(points.size, center, radius)

		self.createVertex2Fiber(fiberSizes)


	# @timeit
	def CPopulateAndDefragment(self, vertexSize, center, radius):
		poolSize = int(vertexSize/self.OT_NODE_CAPACITY*4.4 + 1)

		dt = np.dtype([	('children', np.uint32, (8,)),
						('numberOfElements', np.uint32, (1,)),
						('contentElements', np.uint32, (1,)),
						('radius', np.float32, (3,)),
						('center', np.float32, (3,))])

		self.pool = np.empty(poolSize, dtype=dt)
		self.vertexIdxPool = np.empty(self.vertex.size//3, dtype=np.uint32)

		n = ROIsSegmentationPopulateAndDefragmentPool(
			self.pool,
			poolSize,
			self.vertex,
			self.vertex.size,
			center,
			radius,
			self.vertexIdxPool)

		# n = cfuncs.ROIsSegmentationPopulateAndDefragmentPool(
		# 	self.pool.ctypes.data,
		# 	poolSize,
		# 	self.vertex.ctypes.data,
		# 	self.vertex.size,
		# 	center.ctypes.data,
		# 	radius.ctypes.data,
		# 	self.vertexIdxPool.ctypes.data)

		self.pool = np.array(self.pool[:n], copy=True)


	def createVertex2Fiber(self, fiberSizes):
		self.vertexIdx2FiberIdx = np.empty(self.vertexIdxPool.size, dtype=np.int32)

		offset = 0
		for i in range(len(fiberSizes)):
			self.vertexIdx2FiberIdx[offset:offset+fiberSizes[i]] = i
			offset += fiberSizes[i]

	def queryCollision(self, queryPackage, result):
		ROIsSegmentationQueryOctree(len(queryPackage),
									queryPackage,
									self.pool,
									self.vertexIdxPool,
									self.vertexIdx2FiberIdx,
									self.vertex,
									result,
									result.shape[1])

		# cfuncs.ROIsSegmentationQueryOctree(	len(queryPackage),
		# 									queryPackage.ctypes.data,
		# 									self.pool.ctypes.data,
		# 									self.vertexIdxPool.ctypes.data,
		# 									self.vertexIdx2FiberIdx.ctypes.data,
		# 									self.vertex.ctypes.data,
		# 									result.ctypes.data,
		# 									result.shape[1])



'''
Dictionary is keeping a reference to the segmentations objects... not deleting
'''

import PyQt5
from PyQt5 import QtGui, QtWidgets, uic, QtCore
from importlib_resources import files
from .InPlaceSegmentationDialog import InPlaceSegmentationDialog
from .ROIsSegmentationDialog import ROIsSegmentationDialog
from .AtlasBasedParallelSegmentationDialog import AtlasBasedParallelSegmentationDialog
from .FFClustSegmentationDialog import FFClustSegmentationDialog
from ...Framework.Tools.visualizationEnums import SegmentationTypes, segmentations, mriVisualizations, VisualizationActions, VisualizationObject
from ...Framework.Segmentation.SegmentationHandler import SegmentationHandler

_vot_ui = files('phybers.fibervis.ui').joinpath(
	'visualizationObjectsTool.ui')


def identifyNumberRecursively(item, number):
	root = item

	if item.parent() != None:
		root = identifyNumberRecursively(item.parent(), number)
		number.append(item.parent().indexOfChild(item))
	else:
		return item
	return root


def retrieveItemIteratively(root, index):
	item = root

	for i in index:
		if hasattr(item, 'children'):
			item = item.children[i]
		else:
			item = item[i]

	return item


class ownQTreeWidgetItem(QtWidgets.QTreeWidgetItem):
	def setIdentifier(self, identification):
		self.identifier = identification


class VisDialog(QtWidgets.QDialog):
	''' Window controller for the app. It has function for several action.
	'''

	closed = QtCore.pyqtSignal()
	changeObject = QtCore.pyqtSignal(dict)
	modifySegmentation = QtCore.pyqtSignal(dict)
	selectedObject = QtCore.pyqtSignal(dict)


	def __init__(self, parent):
		super().__init__(parent)
		self.ui = uic.loadUi(str(_vot_ui), self)

		self.ui.setWindowTitle('Visualization Tool')

		# signal when closing
		oldCloseEvent = self.ui.closeEvent
		def newCloseEvent(*args, **kwargs):
			self.closed.emit()
			return oldCloseEvent(*args, **kwargs)
		self.ui.closeEvent = newCloseEvent

		# this 4 variables have to be treated like const
		self.bundleReference = None
		self.meshReference = None
		self.mriReference = None
		self.roisReference = None


		# Context menu
		self.ui.treeWidget.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
		self.ui.treeWidget.customContextMenuRequested.connect(self.prepareMenu)

		# Setting widget
		# Signal for displaying settings
		self.ui.treeWidget.currentItemChanged.connect(self.activeItemChanged)

		self.connectAndConfigRotateGB()
		self.connectAndConfigTranslateGB()
		self.connectAndConfigScaleGB()
		self.ui.resetTransformsButton.clicked.connect(self.resetTransformsItemSelected)
		self.ui.loadMatrixFromFileButton.clicked.connect(self.loadMatrixFromFile)

		self.connectAndConfigBundleGB()
		self.connectAndConfigSliceGB()
		self.connectAndConfigVolumeGB()
		self.configMeshGB()

		# Hide rotate, translate and scale options, show when item is selected
		self.rotateGroupBox.hide()
		self.translateGroupBox.hide()
		self.scaleGroupBox.hide()

		self.bundleGroupBox.hide()
		self.sliceGroupBox.hide()
		self.meshGroupBox.hide()
		self.volumeGroupBox.hide()

		self.ui.resetTransformsButton.hide()

		# Setting windows
		self.windows = dict()

		# segmentation dialogs for identifier
		self.segmentationDialogs = {SegmentationTypes.InPlace : InPlaceSegmentationDialog,
									SegmentationTypes.ROIs : ROIsSegmentationDialog,
									SegmentationTypes.AtlasBased : AtlasBasedParallelSegmentationDialog,
									SegmentationTypes.FFClust : FFClustSegmentationDialog}


	def connectAndConfigRotateGB(self):
		self.ui.rotateAngleLineEdit.returnPressed.connect(self.rotateItemSelected)
		self.ui.rotateXLineEdit.returnPressed.connect(self.rotateItemSelected)
		self.ui.rotateYLineEdit.returnPressed.connect(self.rotateItemSelected)
		self.ui.rotateZLineEdit.returnPressed.connect(self.rotateItemSelected)
		self.ui.rotateButton.clicked.connect(self.rotateItemSelected)

		validator = QtGui.QDoubleValidator(-1000, 1000, 2, self)
		self.ui.rotateAngleLineEdit.setValidator(validator)
		self.ui.rotateXLineEdit.setValidator(validator)
		self.ui.rotateYLineEdit.setValidator(validator)
		self.ui.rotateZLineEdit.setValidator(validator)


	def connectAndConfigTranslateGB(self):
		self.ui.translateXLineEdit.returnPressed.connect(self.translateItemSelected)
		self.ui.translateYLineEdit.returnPressed.connect(self.translateItemSelected)
		self.ui.translateZLineEdit.returnPressed.connect(self.translateItemSelected)
		self.ui.translateButton.clicked.connect(self.translateItemSelected)

		validator = QtGui.QDoubleValidator(-1000, 1000, 2, self)
		self.ui.translateXLineEdit.setValidator(validator)
		self.ui.translateYLineEdit.setValidator(validator)
		self.ui.translateZLineEdit.setValidator(validator)


	def connectAndConfigScaleGB(self):
		self.ui.scaleXLineEdit.returnPressed.connect(self.scaleItemSelected)
		self.ui.scaleYLineEdit.returnPressed.connect(self.scaleItemSelected)
		self.ui.scaleZLineEdit.returnPressed.connect(self.scaleItemSelected)
		self.ui.scaleButton.clicked.connect(self.scaleItemSelected)

		validator = QtGui.QDoubleValidator(-1000, 1000, 2, self)
		self.ui.scaleXLineEdit.setValidator(validator)
		self.ui.scaleYLineEdit.setValidator(validator)
		self.ui.scaleZLineEdit.setValidator(validator)


	def connectAndConfigBundleGB(self):
		self.bundleGBSignals()


	def connectAndConfigSliceGB(self):
		self.sliceGBSignals()

		validator = QtGui.QDoubleValidator(-1000, 1000, 2, self)
		self.ui.axisCXLE.setValidator(validator)
		self.ui.axisCYLE.setValidator(validator)
		self.ui.axisCZLE.setValidator(validator)


	def connectAndConfigVolumeGB(self):
		self.volumeGBSignals()


	def configMeshGB(self, action='connect'):
		getattr(self.ui.drawLinesCheckBox.clicked, action)(self.modifyMeshObject)
		getattr(self.ui.opacitySlider.valueChanged, action)(self.modifyMeshObject)
		getattr(self.ui.setColorButton.clicked, action)(self.modifyMeshObject)
		getattr(self.ui.front2backRadioButton.clicked, action)(self.modifyMeshObject)
		getattr(self.ui.back2frontRadioButton.clicked, action)(self.modifyMeshObject)


	def toggleVisWin(self, isChecked, pos, size):
		@QtCore.pyqtSlot()
		def toggle():

			if isChecked() == True:
				moveX = pos().x()-self.size().width()-5 # 5 pixels away from the window
				moveY = pos().y()-(self.size().height()-size().height())//2

				if moveX < 0:
					moveX = 0
				if moveY < 0:
					moveY = 0
				self.move(moveX,moveY)
				self.show()
			else:
				self.hide()
		return toggle


	def addRoot(self, name, ident):
		root = ownQTreeWidgetItem(self.ui.treeWidget)
		root.setText(0, name)
		self.ui.treeWidget.addTopLevelItem(root)
		root.setExpanded(True)
		root.setIdentifier(ident)
		return root


	def addChild(self, parent, name, ident):
		item = ownQTreeWidgetItem()
		item.setText(0, name)
		parent.addChild(item)
		item.setExpanded(True)
		item.setIdentifier(ident)
		return item


	def removeItem(self, item):
		if item.parent() == None:
			idx = self.ui.treeWidget.indexOfTopLevelItem(item)
			self.ui.treeWidget.takeTopLevelItem(idx)
		else:
			item.parent().removeChild(item)
		del item


	@QtCore.pyqtSlot()
	def updateTree(self):
		self.ui.treeWidget.clear()

		def recursivelyAddChildren(parentItem, parentModel):
			for child in (parentModel.children if hasattr(parentModel, 'children') else parentModel):
				childItem = self.addChild(parentItem, child.fileName, child.identifier)
				if child.identifier == VisualizationObject.Segmentation:
					if not child in self.windows.keys():
						try:
							dialog = self.segmentationDialogs[child.segmentationIdentifier](child, self)
							self.windows[child] = dialog
							dialog.updateObject.connect(self.modifySegmentation)
						except Exception as e:
							print(e)
							raise TypeError('Segmentation type dialog not implemented: ', child.segmentationIdentifier)

					self.windows[child].update()

				recursivelyAddChildren(childItem, child)


		if len(self.bundleReference):
			bundleRoot = self.addRoot('Bundles', VisualizationObject.Bundle)

			recursivelyAddChildren(bundleRoot, self.bundleReference)

		if len(self.meshReference):
			meshRoot = self.addRoot('Meshes', VisualizationObject.Mesh)

			recursivelyAddChildren(meshRoot, self.meshReference)

		if len(self.mriReference):
			mriRoot = self.addRoot('MRIs', VisualizationObject.MRI)

			recursivelyAddChildren(mriRoot, self.mriReference)

		if len(self.roisReference):
			roisRoot = self.addRoot('ROIs', VisualizationObject.ROI)

			recursivelyAddChildren(roisRoot, self.roisReference)


	@QtCore.pyqtSlot(SegmentationHandler)
	def updateSettingWindow(self, segmentationRef):
		self.windows[segmentationRef].updateWindow()


	@QtCore.pyqtSlot(QtCore.QPoint)
	def prepareMenu(self, pos):
		item = self.ui.treeWidget.itemAt(pos)
		if item == None:
			return

		number = []
		root = identifyNumberRecursively(item,number)

		parent = item.parent()
		text = None
		if parent != None:
			text = parent.text(0)

		# Menu
		contextMenu = QtWidgets.QMenu(self.ui.treeWidget)

		# Actions
		deleteAction = contextMenu.addAction('Delete')
		addSegmentationAction = QtWidgets.QMenu('Add segmentation method')
		addMRIVisualizationAction = QtWidgets.QMenu('Add MRI visualization object')
		focusObjectAction = contextMenu.addAction('Focus Camera')
		visibleToggleAction = QtWidgets.QAction('Visible')
		visibleToggleAction.setCheckable(True)

		windowToggleAction = QtWidgets.QAction('Setting window')
		windowToggleAction.setCheckable(True)

		segmentationAction = dict()
		mriVisAction = dict()

		# Segmentation only on bundles
		if item.identifier == VisualizationObject.Bundle and parent != None:
			contextMenu.addMenu(addSegmentationAction)

			for key in segmentations.keys():
				segmentationAction[addSegmentationAction.addAction(key)] = segmentations[key]

		# MRI slicing and volume
		if item.identifier == VisualizationObject.MRI and parent != None:
			contextMenu.addMenu(addMRIVisualizationAction)

			for key in mriVisualizations.keys():
				mriVisAction[addMRIVisualizationAction.addAction(key)] = mriVisualizations[key]


		# Visible only on Objects
		if root.text(0) == 'Bundles':
			reference = self.bundleReference
		elif root.text(0) == 'Meshes':
			reference = self.meshReference
		elif root.text(0) == 'MRIs':
			reference = self.mriReference
		elif root.text(0) == 'ROIs':
			reference = self.roisReference
		else:
			reference = None

		if parent != None:
			visibleToggleAction.setChecked(retrieveItemIteratively(reference,number).drawable)
			contextMenu.addAction(visibleToggleAction)
		if item.identifier == VisualizationObject.Segmentation:
			dialog = self.windows[retrieveItemIteratively(reference,number)]
			windowToggleAction.setChecked(dialog.isVisible())
			contextMenu.addAction(windowToggleAction)

		# test
		testAction = contextMenu.addAction('Test')

		action = contextMenu.exec_(self.ui.treeWidget.mapToGlobal(pos))

		modifyObject = None

		if action == deleteAction:
			if item.identifier == VisualizationObject.Segmentation:
				self.windows.pop(retrieveItemIteratively(reference,number), None)
			modifyObject = self.prepareDictionary(item, VisualizationActions.Delete)
		elif action == visibleToggleAction:
			modifyObject = self.prepareDictionary(item, VisualizationActions.ToggleDrawable)
		elif action == windowToggleAction:
			dialog = self.windows[retrieveItemIteratively(reference, number)]
			dialog.setVisible(not dialog.isVisible())
			return
		elif action == testAction:
			print('testAction')
			return
		elif action in segmentationAction.keys():
			modifyObject = self.prepareDictionary(item, VisualizationActions.AddSegmentation, data=segmentationAction[action])
		elif action in mriVisAction.keys():
			modifyObject = self.prepareDictionary(item, VisualizationActions.AddSegmentation, data=mriVisAction[action])
		elif action == focusObjectAction:
			modifyObject = self.prepareDictionary(item, VisualizationActions.FocusObject)
		else:
			print('Action not atended: ', action)
			return

		self.changeObject.emit(modifyObject)


	def keyPressEvent(self, event):
		if event.key() == QtCore.Qt.Key_Delete:
			item = self.ui.treeWidget.currentItem()

			if item == None:
				return

			modifyObject = self.prepareDictionary(item, VisualizationActions.Delete)

			self.changeObject.emit(modifyObject)

		else:
			event.ignore()
			return

		event.accept()


	def identifyItem(self, item):
		if item.parent() == None:
			identifier = item.identifier
			number = -1
		else:
			number = []
			root = identifyNumberRecursively(item,number)
			identifier = root.identifier

		return identifier, number


	def prepareDictionary(self, item, action, data=None):
		opts = {}
		opts['rootType'], opts['number'] = self.identifyItem(item)
		opts['action'] = action

		if data != None:
			opts['data'] = data

		return opts


	@QtCore.pyqtSlot(QtWidgets.QTreeWidgetItem, QtWidgets.QTreeWidgetItem)
	def activeItemChanged(self, current, previous):
		if current == None:
			self.rotateGroupBox.hide()
			self.translateGroupBox.hide()
			self.scaleGroupBox.hide()
			self.ui.itemSelectedLineEdit.setText('None')

			self.bundleGroupBox.hide()
			self.sliceGroupBox.hide()
			self.meshGroupBox.hide()
			self.volumeGroupBox.hide()
			self.ui.resetTransformsButton.hide()

			activeObject = None

		else:
			self.rotateGroupBox.show()
			self.translateGroupBox.show()
			self.scaleGroupBox.show()
			self.ui.itemSelectedLineEdit.setText(current.text(0))
			self.ui.resetTransformsButton.show()

			number = []
			root = identifyNumberRecursively(current, number)

			# Visible only on Objects
			if root.text(0) == 'Bundles':
				reference = self.bundleReference
			elif root.text(0) == 'Meshes':
				reference = self.meshReference
			elif root.text(0) == 'MRIs':
				reference = self.mriReference
			elif root.text(0) == 'ROIs':
				reference = self.roisReference
			else:
				reference = None

			refObj = retrieveItemIteratively(reference, number)
			activeObject = refObj

			if current.identifier == VisualizationObject.Segmentation:
				print('SEGMENTATION')
			# --------------------BUNDLES--------------------
			if current.identifier == VisualizationObject.Bundle:
				self.bundleGroupBox.show()

				self.bundleGBSignals(action='disconnect')
				if isinstance(refObj, list):
					self.ui.shaderRadioButton_0.setChecked(refObj[0].selectedShader == 0)
					self.ui.shaderRadioButton_1.setChecked(refObj[0].selectedShader == 1)
				else:
					self.ui.shaderRadioButton_0.setChecked(refObj.selectedShader == 0)
					self.ui.shaderRadioButton_1.setChecked(refObj.selectedShader == 1)
				self.bundleGBSignals()

			else:
				self.bundleGroupBox.hide()

			# --------------------MRI SLICE--------------------
			if current.identifier == VisualizationObject.MRISlice:
				self.sliceGroupBox.show()

				self.sliceGBSignals(action='disconnect')
				self.ui.linearInterpCheckBox.setChecked(refObj.linearInterp)
				self.ui.discardPixelsCheckBox.setChecked(refObj.discardValues)
				self.ui.sliceThresholdSlider.setMaximum(int(refObj.max))
				self.ui.sliceThresholdSlider.setValue(int(refObj.threshold))
				self.ui.sliceCurrentThresholdLineEdit.setText(str(refObj.threshold))

				axis = refObj.axis == 1.0
				if axis.sum() == 1:
					if axis[0] == 1.0:
						self.ui.axisXRadioButton.setChecked(True)
					elif axis[1] == 1.0:
						self.ui.axisYRadioButton.setChecked(True)
					elif axis[2] == 1.0:
						self.ui.axisZRadioButton.setChecked(True)
				else:
					self.ui.axisCustomRadioButton.setChecked(True)

				self.axisCXLE.setText(str(refObj.axis[0]))
				self.axisCYLE.setText(str(refObj.axis[1]))
				self.axisCZLE.setText(str(refObj.axis[2]))

				self.slicePos2AxisSlider.setValue(int(refObj.pos2Axis*(self.ui.slicePos2AxisSlider.maximum()+1)))

				self.sliceBrightSlider.setValue(int(refObj.bright*self.sliceBrightSlider.maximum()))
				self.sliceContrastSlider.setValue(int(refObj.contrast*10))
				self.sliceGBSignals()

			else:
				self.sliceGroupBox.hide()

			# --------------------MRI VOLUME--------------------
			if current.identifier == VisualizationObject.MRIVolume:
				self.volumeGBSignals(action='disconnect')
				self.volumeGroupBox.show()

				self.ui.volumeAlphaSlider.setValue(int(refObj.alpha*100))

				self.ui.volumeF2BCheckBox.setChecked(refObj.f2bDrawing)

				self.ui.volumeThresholdSlider.setMaximum(int(refObj.max)*100)
				self.ui.volumeThresholdSlider.setValue(int(refObj.threshold*100))
				self.ui.volumeCurrentThresholdLineEdit.setText(str(refObj.threshold))

				self.volumeGBSignals()

			else:
				self.volumeGroupBox.hide()


			# --------------------MESH--------------------
			if current.identifier == VisualizationObject.Mesh and len(number) != 0:
				self.meshGroupBox.show()

				self.configMeshGB(action='disconnect')
				self.ui.drawLinesCheckBox.setChecked(refObj.drawableLines)
				self.ui.opacitySlider.setValue(int(self.ui.opacitySlider.maximum()*refObj.alpha))

				self.ui.colorWidget.setAutoFillBackground(True)
				self.ui.colorWidget.setPalette(QtGui.QPalette(QtGui.QColor(*(refObj.getRGB256()))))

				if refObj.back2frontSorting:
					self.ui.back2frontRadioButton.setChecked(True)
				else:
					self.ui.front2backRadioButton.setChecked(True)

				self.configMeshGB()

			else:
				self.meshGroupBox.hide()

		self.selectedObject.emit({'current' : activeObject})


	@QtCore.pyqtSlot()
	def rotateItemSelected(self):
		item = self.ui.treeWidget.currentItem()

		if item == None:
			return

		angle = self.ui.rotateAngleLineEdit.text()
		x = self.ui.rotateXLineEdit.text()
		y = self.ui.rotateYLineEdit.text()
		z = self.ui.rotateZLineEdit.text()

		angle = float(angle) if angle != '.' and len(angle) != 0 else 0.0
		x = float(x) if x != '.' and len(x) != 0 else 0.0
		y = float(y) if y != '.' and len(y) != 0 else 0.0
		z = float(z) if z != '.' and len(z) != 0 else 0.0

		axis = [x, y, z]

		if (x**2+y**2+z**2) < 0.0001:
			reply = QtWidgets.QMessageBox.question(self, "Invalid value", 'Axis for rotation was not given.', QtWidgets.QMessageBox.Ok)
			return

		modifyObject = self.prepareDictionary(item, VisualizationActions.Rotate, data=(angle, axis))

		self.changeObject.emit(modifyObject)


	@QtCore.pyqtSlot()
	def resetTransformsItemSelected(self):
		item = self.ui.treeWidget.currentItem()

		if item == None:
			return

		modifyObject = self.prepareDictionary(item, VisualizationActions.ResetTransforms)

		self.changeObject.emit(modifyObject)


	@QtCore.pyqtSlot()
	def loadMatrixFromFile(self):
		item = self.ui.treeWidget.currentItem()

		if item == None:
			return

		matrixFile, fileType = QtWidgets.QFileDialog.getOpenFileNames(self, "Select matrix", "", "matrix (*.npy *.trm *.txt)")

		if len(matrixFile) != 1:
			return

		modifyObject = self.prepareDictionary(item, VisualizationActions.LoadAndApplyMatrix, data=matrixFile[0])

		self.changeObject.emit(modifyObject)


	@QtCore.pyqtSlot()
	def translateItemSelected(self):
		item = self.ui.treeWidget.currentItem()

		if item == None:
			return

		x = self.ui.translateXLineEdit.text()
		y = self.ui.translateYLineEdit.text()
		z = self.ui.translateZLineEdit.text()

		x = float(x) if x != '.' and len(x) != 0 else 0.0
		y = float(y) if y != '.' and len(y) != 0 else 0.0
		z = float(z) if z != '.' and len(z) != 0 else 0.0

		v = [x, y, z]

		modifyObject = self.prepareDictionary(item, VisualizationActions.Translate, data=v)

		self.changeObject.emit(modifyObject)


	@QtCore.pyqtSlot()
	def scaleItemSelected(self):
		item = self.ui.treeWidget.currentItem()

		if item == None:
			print('SHOULDNT BE TRIGGERED')
			return

		x = self.ui.scaleXLineEdit.text()
		y = self.ui.scaleYLineEdit.text()
		z = self.ui.scaleZLineEdit.text()

		x = float(x) if x != '.' and len(x) != 0 else 1.0
		y = float(y) if y != '.' and len(y) != 0 else 1.0
		z = float(z) if z != '.' and len(z) != 0 else 1.0

		v = [x, y, z]

		modifyObject = self.prepareDictionary(item, VisualizationActions.Scale, data=v)

		self.changeObject.emit(modifyObject)


	@QtCore.pyqtSlot()
	def modifyBundleObject(self):
		item = self.ui.treeWidget.currentItem()
		shaderSelected = 0 if self.ui.shaderRadioButton_0.isChecked() else 1

		modifyObject = self.prepareDictionary(item, VisualizationActions.ShaderSelection, data=(shaderSelected))

		self.changeObject.emit(modifyObject)


	@QtCore.pyqtSlot()
	def modifySliceObject(self):
		item = self.ui.treeWidget.currentItem()

		# texture linear interpolation checkbox
		linearInterp = self.ui.linearInterpCheckBox.isChecked()

		# discard values under threshold
		discardValues = self.ui.discardPixelsCheckBox.isChecked()

		# threshold
		threshold = self.ui.sliceThresholdSlider.value()
		self.ui.sliceCurrentThresholdLineEdit.setText(str(threshold))

		# Axis
		if self.ui.axisCustomRadioButton.isChecked():
			x = self.ui.axisCXLE.text()
			y = self.ui.axisCYLE.text()
			z = self.ui.axisCZLE.text()

			x = float(x) if x != '.' and len(x) != 0 else 1.0
			y = float(y) if y != '.' and len(y) != 0 else 1.0
			z = float(z) if z != '.' and len(z) != 0 else 1.0

			if (x**2+y**2+z**2) < 0.0001:
				reply = QtWidgets.QMessageBox.question(self, "Invalid value", 'Axis for plane was not given.', QtWidgets.QMessageBox.Ok)
				return

			axis = [x, y, z]

		else:
			axis = [float(self.ui.axisXRadioButton.isChecked()),
					float(self.ui.axisYRadioButton.isChecked()),
					float(self.ui.axisZRadioButton.isChecked())]

		# pos2Axis slider
		pos2Axis = self.ui.slicePos2AxisSlider.value()/(self.ui.slicePos2AxisSlider.maximum()+1)

		# bright and contrast
		bright = self.ui.sliceBrightSlider.value()/self.ui.sliceBrightSlider.maximum()
		contrast = self.ui.sliceContrastSlider.value()/10

		modifyObject = self.prepareDictionary(item, VisualizationActions.SliceModification, data=(linearInterp, discardValues, threshold, axis, pos2Axis, bright, contrast))

		self.changeObject.emit(modifyObject)

	@QtCore.pyqtSlot()
	def modifyVolumeObject(self):
		item = self.ui.treeWidget.currentItem()

		if self.sender() == self.ui.volumeThresholdSlider or self.sender() == self.ui.volumeAlphaSlider:
			alpha = self.ui.volumeAlphaSlider.value()/100
			threshold = self.ui.volumeThresholdSlider.value()/100

			self.ui.volumeCurrentThresholdLineEdit.setText(str(threshold))

			data = [threshold, alpha]
		else:
			data = self.ui.volumeF2BCheckBox.isChecked()

		modifyObject = self.prepareDictionary(item, VisualizationActions.VolumeModification, data)

		self.changeObject.emit(modifyObject)



	@QtCore.pyqtSlot()
	def modifyMeshObject(self):
		item = self.ui.treeWidget.currentItem()

		sender = self.sender()

		if sender == self.ui.setColorButton:
			color = QtWidgets.QColorDialog.getColor()
			if color.isValid():
				self.ui.colorWidget.setPalette(QtGui.QPalette(color))
				color = (color.red()/255, color.green()/255, color.blue()/255)
			else:
				return

			data = [color]

		elif sender == self.ui.drawLinesCheckBox or sender == self.ui.opacitySlider:

			drawLines = self.ui.drawLinesCheckBox.isChecked()
			opacity = self.ui.opacitySlider.value()/self.ui.opacitySlider.maximum()

			data = (drawLines, opacity)

		elif sender == self.ui.front2backRadioButton:
			data = [False]

		elif sender == self.ui.back2frontRadioButton:
			data = [True]

		else:
			print('Unknown sender for modifyMeshObject in VisDialog:', self.sender())
			return

		modifyObject = self.prepareDictionary(item, VisualizationActions.MeshModification, data=data)

		self.changeObject.emit(modifyObject)


	@QtCore.pyqtSlot(list)
	def objectsChanged(self, objects):
		for dialogs in self.windows.keys():
			try:
				self.windows[dialogs].objectsChanged(objects)
			except:
				pass


	def bundleGBSignals(self, action='connect'):
		getattr(self.ui.shaderRadioButton_0.clicked, action) (self.modifyBundleObject)
		getattr(self.ui.shaderRadioButton_1.clicked, action) (self.modifyBundleObject)

	def sliceGBSignals(self, action='connect'):
		getattr(self.ui.linearInterpCheckBox.clicked, action) (self.modifySliceObject)
		getattr(self.ui.discardPixelsCheckBox.clicked, action) (self.modifySliceObject)
		getattr(self.ui.sliceThresholdSlider.valueChanged, action) (self.modifySliceObject)
		getattr(self.ui.axisXRadioButton.clicked, action) (self.modifySliceObject)
		getattr(self.ui.axisYRadioButton.clicked, action) (self.modifySliceObject)
		getattr(self.ui.axisZRadioButton.clicked, action) (self.modifySliceObject)
		getattr(self.ui.axisCustomRadioButton.clicked, action) (self.modifySliceObject)
		getattr(self.ui.axisCXLE.returnPressed, action) (self.modifySliceObject)
		getattr(self.ui.axisCYLE.returnPressed, action) (self.modifySliceObject)
		getattr(self.ui.axisCZLE.returnPressed, action) (self.modifySliceObject)
		getattr(self.ui.slicePos2AxisSlider.valueChanged, action) (self.modifySliceObject)
		getattr(self.ui.sliceBrightSlider.valueChanged, action) (self.modifySliceObject)
		getattr(self.ui.sliceContrastSlider.valueChanged, action) (self.modifySliceObject)

	def volumeGBSignals(self, action='connect'):
		getattr(self.ui.volumeThresholdSlider.valueChanged, action) (self.modifyVolumeObject)
		getattr(self.ui.volumeAlphaSlider.valueChanged, action) (self.modifyVolumeObject)
		getattr(self.ui.volumeF2BCheckBox.clicked, action) (self.modifyVolumeObject)
		getattr(self.ui.volumeCurrentThresholdLineEdit.returnPressed, action) (self.modifyVolumeObject)



from PyQt5 import QtGui, QtWidgets, uic, QtCore
from importlib_resources import files

_ffcsd_ui = files('phybers.fibervis.ui.Segmentations').joinpath(
	'FFClustSegmentationDialog.ui')

class FFClustSegmentationDialog(QtWidgets.QDialog):
	updateObject = QtCore.pyqtSignal(dict)


	def __init__(self, segmentation, parent):
		super().__init__(parent)
		self.ui = uic.loadUi(str(_ffcsd_ui), self)

		self.segmentationRef = segmentation


		self.ui.alphaSlider.valueChanged.connect(self.alphaChanged)

		self.ui.segmentButton.clicked.connect(self.segmentSubject)
		self.ui.exportBundleButton.clicked.connect(self.exportBundle)
		self.ui.exportCentroidButton.clicked.connect(self.exportCentroid)

		self.show()


	@QtCore.pyqtSlot()
	def segmentSubject(self):
		options = {	'reference' : self.segmentationRef,
					'segmentSubject' : True}

		self.updateObject.emit(options)


	@QtCore.pyqtSlot()
	def alphaChanged(self):
		alpha = self.alphaSlider.value()/(self.alphaSlider.maximum()+1)

		options = {	'reference' : self.segmentationRef,
					'attributes': {'setAlpha' : [alpha]}}

		self.updateObject.emit(options)


	@QtCore.pyqtSlot()
	def exportBundle(self):

		# Gets direction of bundle files
		outFile, fileType = QtWidgets.QFileDialog.getSaveFileName(self, "Save Bundle file", "", "Bundle (*.bundles)")#;;Trk file (*.trk)")

		options = {	'reference' : self.segmentationRef,
					'attributes' : {'exportFile' : [outFile]}}

		self.updateObject.emit(options)


	@QtCore.pyqtSlot()
	def exportCentroid(self):
		# Gets direction of bundle files
		outFile, fileType = QtWidgets.QFileDialog.getSaveFileName(self, "Save Bundle file", "", "Bundle (*.bundles)")#;;Trk file (*.trk)")

		options = {	'reference' : self.segmentationRef,
					'attributes' : {'exportCentroidFile' : [outFile]}}

		self.updateObject.emit(options)





	def __del__(self):
		print('Must clean befor leaving... INPLACESEGMENTATIONDIALOG', type(self))
		segmentationRef = None
		# super.__del__()


	def update(self):
		pass


from OpenGL import GL

class Shader:
	""" Shader class for GLSL programs
	"""
	def __init__(self, vShader,  fShader, gShader=None):
		"""
		Parameters
		----------
		vShader : str or tuple
			String containing shader source code path (or tuple with paths)
			for the vertex shader
		fShader : str or tuple
			String containing shader source code path (or tuple with paths)
			for the fragment shader
		gShader : str or tuple
			String containing shader source code path (or tuple with paths)
			for the geometry shader

		"""
		self.program = GL.glCreateProgram()
		shaders = []

		# vertex shaders
		if type(vShader) is list:
			shaders += [self._addShader(vertex, GL.GL_VERTEX_SHADER)
                    for vertex in vShader]
		else:
			shaders.append(self._addShader(vShader, GL.GL_VERTEX_SHADER))

		# fragment shaders
		if type(fShader) is list:
			shaders += [self._addShader(fragment, GL.GL_FRAGMENT_SHADER)
                    for fragment in fShader]
		else:
			shaders.append(self._addShader(fShader, GL.GL_FRAGMENT_SHADER))

		# geometry shaders
		if type(gShader) is list:
			shaders += [self._addShader(geometry, GL.GL_GEOMETRY_SHADER) for geometry in gShader]
		elif gShader is not None:
			shaders.append(self._addShader(gShader, GL.GL_GEOMETRY_SHADER))

		GL.glLinkProgram(self.program)

		# check if got linked
		if GL.glGetProgramiv(self.program, GL.GL_LINK_STATUS) != GL.GL_TRUE:
			info = GL.glGetProgramInfoLog(self.program)
			GL.glDeleteProgram(self.program)
			for shaders_id in shaders:
				GL.glDeleteShader(shaders_id)
			raise RuntimeError('Error linking program: %s' % (info))

		for shaders_id in shaders:
			GL.glDeleteShader(shaders_id)

	def _addShader(self, source_path, shader_type):
		""" Helper function for compiling a GLSL shader

		Parameters
		----------
		source_path : str
			String containing shader source path

		shader_type : valid OpenGL shader type
			Type of shader to compile

		Returns
		-------
		value : int
			Identifier for shader if compilation is successful

		"""
		shader_id = None
		try:
			shader_id = GL.glCreateShader(shader_type)
			with open(source_path, "r") as file:
				source = file.read()
				GL.glShaderSource(shader_id, source)
			GL.glCompileShader(shader_id)

			if GL.glGetShaderiv(shader_id, GL.GL_COMPILE_STATUS) != GL.GL_TRUE:
				info = GL.glGetShaderInfoLog(shader_id)
				raise RuntimeError('\n\tFile {1}:\n\tShader compilation failed:\n\t\t{0}'.format(info, source_path))

			GL.glAttachShader(self.program, shader_id)
			return shader_id
		except:
			if shader_id is not None:
				GL.glDeleteShader(shader_id)
			raise

	def glGetUniformLocation(self, name):
		""" Helper function to get location of an
		OpenGL uniform variable

		Parameters
		----------
		name : str
			Name of the variable for which location
			is to be returned

		Returns
		-------
		value : int
			Integer describing location

		"""
		uniform = GL.glGetUniformLocation(self.program, name)

		if uniform == -1:
			raise RuntimeError('Uniform not found: {0}'.format(name))
		else:
			return uniform

	def attributeLocation(self, name):
		""" Helper function to get location of an
		OpenGL attribute variable

		Parameters
		----------
		name : str
			Name of the variable for which location is
			to be returned

		Returns
		-------
		value : int
			Integer describing location

		"""
		attribute = GL.glGetAttribLocation(self.program, name)

		if attribute == -1:
			raise RuntimeError('Attribute not found: {0}'.format(name))
		else:
			return attribute

	def glUseProgram(self):
		""" Bipass for the function glUseProgram
		"""

		GL.glUseProgram(self.program)



import time

def timeit(func):

	def timed(*args, **kw):
		ts = time.time()
		result = func(*args, **kw)
		te = time.time()

		print ('%r (%r, %r) %2.4f sec' % (func.__name__, args, kw, te-ts))
		return result

	return timed

def averageTimeit(func):
	n=50

	def timed(*args, **kw):
		time_average = 0
		for i in range(n):
			print('test n: ', i)
			ts = time.time()
			result = func(*args, **kw)
			te = time.time()

			time_average += te-ts

		time_average /= n

		
		print ('%r (%r, %r) %d %2.4f sec' % (func.__name__, args, kw, n, time_average))
		return result

	return timed


import numpy as np

def bytes2MB(bytes):
	return bytes/1024

def bytes2GB(bytes):
	return bytes/(1024*1024)

def findIntegersMultiplierFor(n, under=1):
	for i in range(under):
		tmpN = n+i
		m1 = 1
		m2 = n

		while m1<=m2:
			if tmpN%m1 == 0 and m2<=under:
				return (np.int32(m2), np.int32(m1))
			m1 += 1
			m2 = tmpN//m1

	return None


from PyQt5 import QtGui, QtWidgets, uic, QtCore
from importlib_resources import files

_abpsd_ui = files('phybers.fibervis.ui.Segmentations').joinpath(
	'AtlasBasedParallelSegmentationDialog.ui')

class AtlasBasedParallelSegmentationDialog(QtWidgets.QDialog):
	updateObject = QtCore.pyqtSignal(dict)

	def __init__(self, segmentation, parent):
		super().__init__(parent)
		self.ui = uic.loadUi(str(_abpsd_ui), self)

		self.segmentationRef = segmentation

		self.bundleNames = None
		self.bundleSize = None
		self.states = []

		self.ui.tableWidget.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
		self.ui.tableWidget.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
		self.ui.tableWidget.setColumnWidth(0, 150)
		self.ui.tableWidget.setColumnWidth(2, 50)
		self.ui.tableWidget.horizontalHeader().setStretchLastSection(True)

		self.ui.tableWidget.cellClicked.connect(self.toggleCheck)

		self.ui.checkSegmentAll.clicked.connect(self.checkSegmentAllBundles)

		self.ui.alphaSlider.valueChanged.connect(self.alphaChanged)

		self.ui.addBundlesButton.clicked.connect(self.addBundles)
		self.ui.removeBundlesButton.clicked.connect(self.removeBundles)
		self.ui.applyMatrixButton.clicked.connect(self.applyMatrix)

		self.ui.addThresholdFileButton.clicked.connect(self.addThresholdFile)

		self.ui.filterLineEdit.textChanged.connect(self.filter)

		self.ui.segmentButton.clicked.connect(self.segmentSubject)
		self.ui.exportBundleButton.clicked.connect(self.exportBundle)

		self.show()


	@QtCore.pyqtSlot()
	def segmentSubject(self):
		options = {	'reference' : self.segmentationRef,
					'segmentSubject' : True}

		self.updateObject.emit(options)


	def checkSegmentAllBundles(self):
		print('checkSegmentAll')


	def addThresholdFile(self):
		# Gets direction of bundle files
		thresholdFile, fileType = QtWidgets.QFileDialog.getOpenFileNames(self, "Select threshold file", "", "Txt (*.txt)")

		if len(thresholdFile) != 1:
			return

		options = {	'reference' : self.segmentationRef,
					'attributes': {'updateAtlas' : [None, None, thresholdFile[0]]},
					'updateViewingObjects' : True,
					'updateSettingWindow' : True}

		self.updateObject.emit(options)


	def addBundles(self):
		# Gets direction of bundle files
		bundleFiles, fileType = QtWidgets.QFileDialog.getOpenFileNames(self, "Select Bundle", "", "Bundles (*.bundles *.trk)")

		if len(bundleFiles) == 0:
			return

		options = {	'reference' : self.segmentationRef,
					'attributes': {'updateAtlas' : [bundleFiles]},
					'updateViewingObjects' : True,
					'updateSettingWindow' : True}

		self.updateObject.emit(options)


	def removeBundles(self):
		pass


	@QtCore.pyqtSlot(int)
	def toggleCheck(self, row):
		tmpItem = self.ui.tableWidget.item(row,0)
		tmpItem.setCheckState(not tmpItem.checkState())


	# @QtCore.pyqtSlot()
	# def checkSegmentAll(self):
	# 	self.ui.tableWidget.cellChanged.disconnect(self.toggleStates)

	# 	visualize = self.ui.checkSegmentAll.isChecked()

	# 	for i in range(self.ui.tableWidget.rowCount()):
	# 		item = self.ui.tableWidget.item(i,0)
	# 		item.setCheckState(visualize)
	# 		self.states[self.bundleNames.index(item.text())] = visualize

	# 	if self.ui.checkRender.isChecked():
	# 		self.updateObject.emit(self.returnDict())

	# 	self.ui.tableWidget.cellChanged.connect(self.toggleStates)


	# TMP!!!!!!
	@QtCore.pyqtSlot(str)
	def filter(self, text):
		self.ui.tableWidget.cellChanged.disconnect(self.toggleStates)

		self.ui.tableWidget.clearContents()
		self.ui.tableWidget.setRowCount(0)

		row = 0
		for i in range(len(self.bundleNames)):
			if self.bundleNames[i].find(text) != -1:
				self.ui.tableWidget.insertRow(row)
				item = QtWidgets.QTableWidgetItem(self.bundleNames[i])
				item.setCheckState(self.states[i])
				self.ui.tableWidget.setItem(row, 0, item)

				self.ui.tableWidget.setItem(row, 1, QtWidgets.QTableWidgetItem(self.bundleSize[i]))
				self.ui.tableWidget.item(row,1).setTextAlignment(QtCore.Qt.AlignRight)
				row = row + 1

		checkStateN = 0

		for i in range(self.ui.tableWidget.rowCount()):
			if self.ui.tableWidget.item(i,0).checkState():
				checkStateN += 1

		if checkStateN == self.ui.tableWidget.rowCount():
			self.ui.checkSegmentAll.setChecked(True)
		elif checkStateN == 0:
			self.ui.checkSegmentAll.setChecked(False)

		self.ui.tableWidget.cellChanged.connect(self.toggleStates)


	@QtCore.pyqtSlot(int)
	def toggleStates(self, row):
		item = self.ui.tableWidget.item(row,0)
		tmp = self.bundleNames.index(item.text())
		self.states[tmp] = bool(item.checkState())

		tmp = 0

		for i in range(self.ui.tableWidget.rowCount()):
			tmp += int(self.ui.tableWidget.item(i,0).checkState())

		if tmp == self.ui.tableWidget.rowCount():
			self.ui.checkSegmentAll.setChecked(True)
		else:
			self.ui.checkSegmentAll.setChecked(False)

		if self.ui.checkRender.isChecked():
			self.updateObject.emit(self.returnDict())


	@QtCore.pyqtSlot()
	def exportBundle(self):

		# Gets direction of bundle files
		outFile, fileType = QtWidgets.QFileDialog.getSaveFileName(self, "Save Bundle file", "", "Bundle (*.bundles)")#;;Trk file (*.trk)")

		options = {	'reference' : self.segmentationRef,
					'attributes' : {'exportFile' : [outFile]}}

		self.updateObject.emit(options)


	@QtCore.pyqtSlot()
	def alphaChanged(self):
		alpha = self.alphaSlider.value()/(self.alphaSlider.maximum()+1)

		options = {	'reference' : self.segmentationRef,
					'attributes': {'setAlpha' : [alpha]}}

		self.updateObject.emit(options)


	@QtCore.pyqtSlot()
	def applyMatrix(self):
		matrixFile, fileType = QtWidgets.QFileDialog.getOpenFileNames(self, "Select matrix", "", "matrix (*.npy *.trm)")

		if len(matrixFile) != 1:
			return

		options = {	'reference' : self.segmentationRef,
					'attributes': {'updateAtlas' : [None, None, None, matrixFile[0]]},
					'updateViewingObjects' : True,
					'updateSettingWindow' : False}

		self.updateObject.emit(options)


	def updateWindow(self):

		validTableData = self.segmentationRef.atlas.getValidBundles()
		invalidTableData = self.segmentationRef.atlas.getInvalidBundles()

		self.ui.tableWidget.clearContents()
		self.ui.tableWidget.setRowCount(0)

		for i in range(len(validTableData)):
			self.ui.tableWidget.insertRow(i)
			data = validTableData[i]

			name = QtWidgets.QTableWidgetItem(data[0])
			nfiber = QtWidgets.QTableWidgetItem(str(data[1]))
			hie = QtWidgets.QTableWidgetItem()
			thr = QtWidgets.QTableWidgetItem(str(data[3]))

			hie.setBackground(QtGui.QBrush(QtGui.QColor(
													int(data[2][0]*255),
													int(data[2][1]*255),
													int(data[2][2]*255))))

			self.ui.tableWidget.setItem(i, 0, name)
			self.ui.tableWidget.setItem(i, 1, nfiber)
			self.ui.tableWidget.setItem(i, 2, hie)
			self.ui.tableWidget.setItem(i, 3, thr)

		for i in range(len(invalidTableData)):
			j = i+len(validTableData)
			self.ui.tableWidget.insertRow(j)
			data = invalidTableData[i]

			name = QtWidgets.QTableWidgetItem(data[0])
			nfiber = QtWidgets.QTableWidgetItem(str(data[1]))
			hie = QtWidgets.QTableWidgetItem()
			thr = QtWidgets.QTableWidgetItem(str(data[3]))

			name.setBackground(QtGui.QBrush(QtGui.QColor(255,0,0)))
			nfiber.setBackground(QtGui.QBrush(QtGui.QColor(255,0,0)))
			hie.setBackground(QtGui.QBrush(QtGui.QColor(0,0,0)))
			thr.setBackground(QtGui.QBrush(QtGui.QColor(255,0,0)))

			self.ui.tableWidget.setItem(j, 0, name)
			self.ui.tableWidget.setItem(j, 1, nfiber)
			self.ui.tableWidget.setItem(j, 2, hie)
			self.ui.tableWidget.setItem(j, 3, thr)


	def returnDict(self):
		return {'reference' : self.segmentationRef,
				'attributes' : {'setPercentage' : [self.ui.percentSlider.value()],
								'setStates' : [self.states]},
				'segmentSubject' : True}


	def __del__(self):
		print('Must clean befor leaving... INPLACESEGMENTATIONDIALOG', type(self))
		segmentationRef = None
		# super.__del__()



from enum import Enum

class VisualizationObject(Enum):
	NotDefined = -1
	Bundle = 0
	Mesh = 1
	Atlas = 7
	MRI = 2
	Segmentation = 3
	MRIVolume = 4
	MRISlice = 5
	ROI = 6

class ROIType(Enum):
	Sphere = 0
	Aabb = 1
	Obb = 2
	Plane = 3

class VisualizationActions(Enum):
	Delete = 0
	ToggleDrawable = 1
	Rotate = 2
	Translate = 3
	Scale = 4
	AddSegmentation = 5
	SliceModification = 6
	VolumeModification = 7
	MeshModification = 8
	ResetTransforms = 9
	LoadAndApplyMatrix = 10
	ShaderSelection = 11
	FocusObject = 12

class SegmentationTypes(Enum):
	InPlace = 0
	ROIs = 1
	AtlasBased = 2
	FFClust = 3
	Test = 23

segmentations = {	'In place segmentation' : SegmentationTypes.InPlace,
					'Roi segmentation' : SegmentationTypes.ROIs,
					'Euclidean distance segmentation' : SegmentationTypes.AtlasBased,
					'FFClust' : SegmentationTypes.FFClust,
					'Test segmentation' : SegmentationTypes.Test}

mriVisualizations = {	'Volume render' : VisualizationObject.MRIVolume,
						'Slice' : VisualizationObject.MRISlice}


''' Base class for every drawable object
'''

from functools import wraps
from OpenGL import GL
import numpy as np
from .Tools import glm
from .Tools.visualizationEnums import VisualizationObject


fSize = np.dtype(np.float32).itemsize
iSize = np.dtype(np.int32).itemsize
uSize = np.dtype(np.uint32).itemsize


def propagateToChildren(func):
	""" Decorator for propagate the drawing function to the object's children.
	"""
	@wraps(func)
	def wrapper(*args, **kwargs):
		funcsName = func.__name__

		for child in args[0].children:
			getattr(child, funcsName)(*args[:-1], **kwargs)

		return func(*args, **kwargs)

	return wrapper

def drawable(func):
	""" Decorator for checking the drawable flag state from the object
	that calls the draw function from any VisualizationBaseObject inherited class.
	"""
	@wraps(func)
	def wrapper(*args, **kwargs):
		if args[0].drawable is True:
			return func(*args, **kwargs)
	return wrapper


def config(func):
	""" Decorator for draw function in any VisualizationBaseObject inherited class.
	Configures the using program and the desired vao
	and unlinks the vao when it's finished
	"""
	@wraps(func)
	def wrapper(*args, **kwargs):
		obj = args[0]
		obj.shader[obj.selectedShader].glUseProgram()
		obj.loadUniforms()
		if type(obj.vao) is not np.ndarray:
			GL.glBindVertexArray(obj.vao)
		else:
			GL.glBindVertexArray(obj.vao[obj.selectedShader])
		func(*args, **kwargs)	# the draw function
		GL.glBindVertexArray(0)
	return wrapper


class VisualizationBaseObject:
	def __init__(self, parent=None, shaderDic=None):

		# Parent and children
		self.parent = parent
		self.children = []
		self.identifier = VisualizationObject.NotDefined

		# OpenGL buffer references
		self.vbo = None
		self.ebo = None
		self.vao = None

		self.texture = None

		# OpenGL shader program handler
		if isinstance(shaderDic, dict):
			self.shader = shaderDic[type(self)]
		else:
			self.shader = shaderDic
		self.nShader = 1

		# Matrixs
		self.model = glm.identity()

		self.inverseModel = glm.identity()

		self.rotationMat = glm.identity()
		self.translateMat = glm.identity()
		self.scaleMat = glm.identity()

		# Flags
		self.drawable = False
		self.clean = False
		self.selectedShader = 0

		self.boundingbox = None


	def cleanOpenGL(self):
		print('Must clean befor leaving... ', type(self))


	def __del__(self):
		if not self.clean:
			self.cleanOpenGL()


	@propagateToChildren
	def clearRef(self):
		self.parent = None
		self.children.clear()


	def draw(self):
		pass


	def setCamera(self, newCamera):
		pass

	def get_center(self):
		raise NotImplementedError

	def get_size(self):
		raise NotImplementedError

	def loadUniforms(self):

		GL.glUniformMatrix4fv(self.shader[self.selectedShader].glGetUniformLocation(
			"M"), 1, GL.GL_TRUE, self.model.getA())


	def remove(self, item):
		self.children.remove(item)


	def rotate(self, center, angle, axis):
		axis = (np.append(axis, 1) * self.rotationMat).getA()[0, :3]
		rotate = glm.translateMatrix(center) * glm.rotateMatrix(angle, axis) * glm.translateMatrix(-center)

		self.rotationMat *= rotate

		self.model = self.translateMat*self.rotationMat*self.scaleMat
		self.inverseModel = np.linalg.inv(self.model)

	def spin(self, angle, axis):
		self.rotate(self.get_center(), angle, axis)


	def translate(self, vec):
		self.translateMat = glm.translateMatrix(vec)

		self.model = self.translateMat*self.rotationMat*self.scaleMat
		self.inverseModel = np.linalg.inv(self.model)


	def stackTranslate(self, vec):
		newTranslateMat = glm.translateMatrix(vec)

		self.translateMat = newTranslateMat*self.translateMat

		self.model = self.translateMat*self.rotationMat*self.scaleMat
		self.inverseModel = np.linalg.inv(self.model)


	def scale(self, vec, center):
		self.scaleMat = glm.translateMatrix(center)*glm.scaleMatrix(vec)*glm.translateMatrix(-center)

		self.model = self.translateMat*self.rotationMat*self.scaleMat
		self.inverseModel = np.linalg.inv(self.model)


	def stackScale(self, vec, center):
		newScaleMat = glm.translateMatrix(center)*glm.scaleMatrix(vec)*glm.translateMatrix(-center)

		self.scaleMat = newScaleMat*self.scaleMat

		self.model = self.translateMat*self.rotationMat*self.scaleMat
		self.inverseModel = np.linalg.inv(self.model)


	def resetModel(self):
		self.scaleMat = glm.identity()

		self.translateMat = glm.identity()

		self.rotationMat = glm.identity()

		self.model = glm.identity()
		self.inverseModel = glm.identity()


	def setSelectedShader(self, newShaderId):
		if newShaderId < self.nShader:
			self.selectedShader = newShaderId
		else:
			raise TypeError('Unsupported shaderID, maximun is ', self.nShader, ' got ', newShaderId, '.')

	def setDrawBB(self, boolean):
		if self.boundingbox is not None:
			self.boundingbox.drawable = boolean
